package srctypes

const (
	DockerImageScheme = "docker-image"
	GitScheme         = "git"
	LocalScheme       = "local"
	HTTPScheme        = "http"
	HTTPSScheme       = "https"
	OCIScheme         = "oci-layout"
)
