// Code generated by protoc-gen-go. DO NOT EDIT.
// versions:
// 	protoc-gen-go v1.36.9
// 	protoc        v3.11.4
// source: github.com/moby/buildkit/session/auth/auth.proto

package auth

import (
	protoreflect "google.golang.org/protobuf/reflect/protoreflect"
	protoimpl "google.golang.org/protobuf/runtime/protoimpl"
	reflect "reflect"
	sync "sync"
	unsafe "unsafe"
)

const (
	// Verify that this generated code is sufficiently up-to-date.
	_ = protoimpl.EnforceVersion(20 - protoimpl.MinVersion)
	// Verify that runtime/protoimpl is sufficiently up-to-date.
	_ = protoimpl.EnforceVersion(protoimpl.MaxVersion - 20)
)

type CredentialsRequest struct {
	state         protoimpl.MessageState `protogen:"open.v1"`
	Host          string                 `protobuf:"bytes,1,opt,name=Host,proto3" json:"Host,omitempty"`
	unknownFields protoimpl.UnknownFields
	sizeCache     protoimpl.SizeCache
}

func (x *CredentialsRequest) Reset() {
	*x = CredentialsRequest{}
	mi := &file_github_com_moby_buildkit_session_auth_auth_proto_msgTypes[0]
	ms := protoimpl.X.MessageStateOf(protoimpl.Pointer(x))
	ms.StoreMessageInfo(mi)
}

func (x *CredentialsRequest) String() string {
	return protoimpl.X.MessageStringOf(x)
}

func (*CredentialsRequest) ProtoMessage() {}

func (x *CredentialsRequest) ProtoReflect() protoreflect.Message {
	mi := &file_github_com_moby_buildkit_session_auth_auth_proto_msgTypes[0]
	if x != nil {
		ms := protoimpl.X.MessageStateOf(protoimpl.Pointer(x))
		if ms.LoadMessageInfo() == nil {
			ms.StoreMessageInfo(mi)
		}
		return ms
	}
	return mi.MessageOf(x)
}

// Deprecated: Use CredentialsRequest.ProtoReflect.Descriptor instead.
func (*CredentialsRequest) Descriptor() ([]byte, []int) {
	return file_github_com_moby_buildkit_session_auth_auth_proto_rawDescGZIP(), []int{0}
}

func (x *CredentialsRequest) GetHost() string {
	if x != nil {
		return x.Host
	}
	return ""
}

type CredentialsResponse struct {
	state         protoimpl.MessageState `protogen:"open.v1"`
	Username      string                 `protobuf:"bytes,1,opt,name=Username,proto3" json:"Username,omitempty"`
	Secret        string                 `protobuf:"bytes,2,opt,name=Secret,proto3" json:"Secret,omitempty"`
	unknownFields protoimpl.UnknownFields
	sizeCache     protoimpl.SizeCache
}

func (x *CredentialsResponse) Reset() {
	*x = CredentialsResponse{}
	mi := &file_github_com_moby_buildkit_session_auth_auth_proto_msgTypes[1]
	ms := protoimpl.X.MessageStateOf(protoimpl.Pointer(x))
	ms.StoreMessageInfo(mi)
}

func (x *CredentialsResponse) String() string {
	return protoimpl.X.MessageStringOf(x)
}

func (*CredentialsResponse) ProtoMessage() {}

func (x *CredentialsResponse) ProtoReflect() protoreflect.Message {
	mi := &file_github_com_moby_buildkit_session_auth_auth_proto_msgTypes[1]
	if x != nil {
		ms := protoimpl.X.MessageStateOf(protoimpl.Pointer(x))
		if ms.LoadMessageInfo() == nil {
			ms.StoreMessageInfo(mi)
		}
		return ms
	}
	return mi.MessageOf(x)
}

// Deprecated: Use CredentialsResponse.ProtoReflect.Descriptor instead.
func (*CredentialsResponse) Descriptor() ([]byte, []int) {
	return file_github_com_moby_buildkit_session_auth_auth_proto_rawDescGZIP(), []int{1}
}

func (x *CredentialsResponse) GetUsername() string {
	if x != nil {
		return x.Username
	}
	return ""
}

func (x *CredentialsResponse) GetSecret() string {
	if x != nil {
		return x.Secret
	}
	return ""
}

type FetchTokenRequest struct {
	state         protoimpl.MessageState `protogen:"open.v1"`
	ClientID      string                 `protobuf:"bytes,1,opt,name=ClientID,proto3" json:"ClientID,omitempty"`
	Host          string                 `protobuf:"bytes,2,opt,name=Host,proto3" json:"Host,omitempty"`
	Realm         string                 `protobuf:"bytes,3,opt,name=Realm,proto3" json:"Realm,omitempty"`
	Service       string                 `protobuf:"bytes,4,opt,name=Service,proto3" json:"Service,omitempty"`
	Scopes        []string               `protobuf:"bytes,5,rep,name=Scopes,proto3" json:"Scopes,omitempty"`
	unknownFields protoimpl.UnknownFields
	sizeCache     protoimpl.SizeCache
}

func (x *FetchTokenRequest) Reset() {
	*x = FetchTokenRequest{}
	mi := &file_github_com_moby_buildkit_session_auth_auth_proto_msgTypes[2]
	ms := protoimpl.X.MessageStateOf(protoimpl.Pointer(x))
	ms.StoreMessageInfo(mi)
}

func (x *FetchTokenRequest) String() string {
	return protoimpl.X.MessageStringOf(x)
}

func (*FetchTokenRequest) ProtoMessage() {}

func (x *FetchTokenRequest) ProtoReflect() protoreflect.Message {
	mi := &file_github_com_moby_buildkit_session_auth_auth_proto_msgTypes[2]
	if x != nil {
		ms := protoimpl.X.MessageStateOf(protoimpl.Pointer(x))
		if ms.LoadMessageInfo() == nil {
			ms.StoreMessageInfo(mi)
		}
		return ms
	}
	return mi.MessageOf(x)
}

// Deprecated: Use FetchTokenRequest.ProtoReflect.Descriptor instead.
func (*FetchTokenRequest) Descriptor() ([]byte, []int) {
	return file_github_com_moby_buildkit_session_auth_auth_proto_rawDescGZIP(), []int{2}
}

func (x *FetchTokenRequest) GetClientID() string {
	if x != nil {
		return x.ClientID
	}
	return ""
}

func (x *FetchTokenRequest) GetHost() string {
	if x != nil {
		return x.Host
	}
	return ""
}

func (x *FetchTokenRequest) GetRealm() string {
	if x != nil {
		return x.Realm
	}
	return ""
}

func (x *FetchTokenRequest) GetService() string {
	if x != nil {
		return x.Service
	}
	return ""
}

func (x *FetchTokenRequest) GetScopes() []string {
	if x != nil {
		return x.Scopes
	}
	return nil
}

type FetchTokenResponse struct {
	state         protoimpl.MessageState `protogen:"open.v1"`
	Token         string                 `protobuf:"bytes,1,opt,name=Token,proto3" json:"Token,omitempty"`
	ExpiresIn     int64                  `protobuf:"varint,2,opt,name=ExpiresIn,proto3" json:"ExpiresIn,omitempty"` // seconds
	IssuedAt      int64                  `protobuf:"varint,3,opt,name=IssuedAt,proto3" json:"IssuedAt,omitempty"`   // timestamp
	unknownFields protoimpl.UnknownFields
	sizeCache     protoimpl.SizeCache
}

func (x *FetchTokenResponse) Reset() {
	*x = FetchTokenResponse{}
	mi := &file_github_com_moby_buildkit_session_auth_auth_proto_msgTypes[3]
	ms := protoimpl.X.MessageStateOf(protoimpl.Pointer(x))
	ms.StoreMessageInfo(mi)
}

func (x *FetchTokenResponse) String() string {
	return protoimpl.X.MessageStringOf(x)
}

func (*FetchTokenResponse) ProtoMessage() {}

func (x *FetchTokenResponse) ProtoReflect() protoreflect.Message {
	mi := &file_github_com_moby_buildkit_session_auth_auth_proto_msgTypes[3]
	if x != nil {
		ms := protoimpl.X.MessageStateOf(protoimpl.Pointer(x))
		if ms.LoadMessageInfo() == nil {
			ms.StoreMessageInfo(mi)
		}
		return ms
	}
	return mi.MessageOf(x)
}

// Deprecated: Use FetchTokenResponse.ProtoReflect.Descriptor instead.
func (*FetchTokenResponse) Descriptor() ([]byte, []int) {
	return file_github_com_moby_buildkit_session_auth_auth_proto_rawDescGZIP(), []int{3}
}

func (x *FetchTokenResponse) GetToken() string {
	if x != nil {
		return x.Token
	}
	return ""
}

func (x *FetchTokenResponse) GetExpiresIn() int64 {
	if x != nil {
		return x.ExpiresIn
	}
	return 0
}

func (x *FetchTokenResponse) GetIssuedAt() int64 {
	if x != nil {
		return x.IssuedAt
	}
	return 0
}

type GetTokenAuthorityRequest struct {
	state         protoimpl.MessageState `protogen:"open.v1"`
	Host          string                 `protobuf:"bytes,1,opt,name=Host,proto3" json:"Host,omitempty"`
	Salt          []byte                 `protobuf:"bytes,2,opt,name=Salt,proto3" json:"Salt,omitempty"`
	unknownFields protoimpl.UnknownFields
	sizeCache     protoimpl.SizeCache
}

func (x *GetTokenAuthorityRequest) Reset() {
	*x = GetTokenAuthorityRequest{}
	mi := &file_github_com_moby_buildkit_session_auth_auth_proto_msgTypes[4]
	ms := protoimpl.X.MessageStateOf(protoimpl.Pointer(x))
	ms.StoreMessageInfo(mi)
}

func (x *GetTokenAuthorityRequest) String() string {
	return protoimpl.X.MessageStringOf(x)
}

func (*GetTokenAuthorityRequest) ProtoMessage() {}

func (x *GetTokenAuthorityRequest) ProtoReflect() protoreflect.Message {
	mi := &file_github_com_moby_buildkit_session_auth_auth_proto_msgTypes[4]
	if x != nil {
		ms := protoimpl.X.MessageStateOf(protoimpl.Pointer(x))
		if ms.LoadMessageInfo() == nil {
			ms.StoreMessageInfo(mi)
		}
		return ms
	}
	return mi.MessageOf(x)
}

// Deprecated: Use GetTokenAuthorityRequest.ProtoReflect.Descriptor instead.
func (*GetTokenAuthorityRequest) Descriptor() ([]byte, []int) {
	return file_github_com_moby_buildkit_session_auth_auth_proto_rawDescGZIP(), []int{4}
}

func (x *GetTokenAuthorityRequest) GetHost() string {
	if x != nil {
		return x.Host
	}
	return ""
}

func (x *GetTokenAuthorityRequest) GetSalt() []byte {
	if x != nil {
		return x.Salt
	}
	return nil
}

type GetTokenAuthorityResponse struct {
	state         protoimpl.MessageState `protogen:"open.v1"`
	PublicKey     []byte                 `protobuf:"bytes,1,opt,name=PublicKey,proto3" json:"PublicKey,omitempty"`
	unknownFields protoimpl.UnknownFields
	sizeCache     protoimpl.SizeCache
}

func (x *GetTokenAuthorityResponse) Reset() {
	*x = GetTokenAuthorityResponse{}
	mi := &file_github_com_moby_buildkit_session_auth_auth_proto_msgTypes[5]
	ms := protoimpl.X.MessageStateOf(protoimpl.Pointer(x))
	ms.StoreMessageInfo(mi)
}

func (x *GetTokenAuthorityResponse) String() string {
	return protoimpl.X.MessageStringOf(x)
}

func (*GetTokenAuthorityResponse) ProtoMessage() {}

func (x *GetTokenAuthorityResponse) ProtoReflect() protoreflect.Message {
	mi := &file_github_com_moby_buildkit_session_auth_auth_proto_msgTypes[5]
	if x != nil {
		ms := protoimpl.X.MessageStateOf(protoimpl.Pointer(x))
		if ms.LoadMessageInfo() == nil {
			ms.StoreMessageInfo(mi)
		}
		return ms
	}
	return mi.MessageOf(x)
}

// Deprecated: Use GetTokenAuthorityResponse.ProtoReflect.Descriptor instead.
func (*GetTokenAuthorityResponse) Descriptor() ([]byte, []int) {
	return file_github_com_moby_buildkit_session_auth_auth_proto_rawDescGZIP(), []int{5}
}

func (x *GetTokenAuthorityResponse) GetPublicKey() []byte {
	if x != nil {
		return x.PublicKey
	}
	return nil
}

type VerifyTokenAuthorityRequest struct {
	state         protoimpl.MessageState `protogen:"open.v1"`
	Host          string                 `protobuf:"bytes,1,opt,name=Host,proto3" json:"Host,omitempty"`
	Payload       []byte                 `protobuf:"bytes,2,opt,name=Payload,proto3" json:"Payload,omitempty"`
	Salt          []byte                 `protobuf:"bytes,3,opt,name=Salt,proto3" json:"Salt,omitempty"`
	unknownFields protoimpl.UnknownFields
	sizeCache     protoimpl.SizeCache
}

func (x *VerifyTokenAuthorityRequest) Reset() {
	*x = VerifyTokenAuthorityRequest{}
	mi := &file_github_com_moby_buildkit_session_auth_auth_proto_msgTypes[6]
	ms := protoimpl.X.MessageStateOf(protoimpl.Pointer(x))
	ms.StoreMessageInfo(mi)
}

func (x *VerifyTokenAuthorityRequest) String() string {
	return protoimpl.X.MessageStringOf(x)
}

func (*VerifyTokenAuthorityRequest) ProtoMessage() {}

func (x *VerifyTokenAuthorityRequest) ProtoReflect() protoreflect.Message {
	mi := &file_github_com_moby_buildkit_session_auth_auth_proto_msgTypes[6]
	if x != nil {
		ms := protoimpl.X.MessageStateOf(protoimpl.Pointer(x))
		if ms.LoadMessageInfo() == nil {
			ms.StoreMessageInfo(mi)
		}
		return ms
	}
	return mi.MessageOf(x)
}

// Deprecated: Use VerifyTokenAuthorityRequest.ProtoReflect.Descriptor instead.
func (*VerifyTokenAuthorityRequest) Descriptor() ([]byte, []int) {
	return file_github_com_moby_buildkit_session_auth_auth_proto_rawDescGZIP(), []int{6}
}

func (x *VerifyTokenAuthorityRequest) GetHost() string {
	if x != nil {
		return x.Host
	}
	return ""
}

func (x *VerifyTokenAuthorityRequest) GetPayload() []byte {
	if x != nil {
		return x.Payload
	}
	return nil
}

func (x *VerifyTokenAuthorityRequest) GetSalt() []byte {
	if x != nil {
		return x.Salt
	}
	return nil
}

type VerifyTokenAuthorityResponse struct {
	state         protoimpl.MessageState `protogen:"open.v1"`
	Signed        []byte                 `protobuf:"bytes,1,opt,name=Signed,proto3" json:"Signed,omitempty"`
	unknownFields protoimpl.UnknownFields
	sizeCache     protoimpl.SizeCache
}

func (x *VerifyTokenAuthorityResponse) Reset() {
	*x = VerifyTokenAuthorityResponse{}
	mi := &file_github_com_moby_buildkit_session_auth_auth_proto_msgTypes[7]
	ms := protoimpl.X.MessageStateOf(protoimpl.Pointer(x))
	ms.StoreMessageInfo(mi)
}

func (x *VerifyTokenAuthorityResponse) String() string {
	return protoimpl.X.MessageStringOf(x)
}

func (*VerifyTokenAuthorityResponse) ProtoMessage() {}

func (x *VerifyTokenAuthorityResponse) ProtoReflect() protoreflect.Message {
	mi := &file_github_com_moby_buildkit_session_auth_auth_proto_msgTypes[7]
	if x != nil {
		ms := protoimpl.X.MessageStateOf(protoimpl.Pointer(x))
		if ms.LoadMessageInfo() == nil {
			ms.StoreMessageInfo(mi)
		}
		return ms
	}
	return mi.MessageOf(x)
}

// Deprecated: Use VerifyTokenAuthorityResponse.ProtoReflect.Descriptor instead.
func (*VerifyTokenAuthorityResponse) Descriptor() ([]byte, []int) {
	return file_github_com_moby_buildkit_session_auth_auth_proto_rawDescGZIP(), []int{7}
}

func (x *VerifyTokenAuthorityResponse) GetSigned() []byte {
	if x != nil {
		return x.Signed
	}
	return nil
}

var File_github_com_moby_buildkit_session_auth_auth_proto protoreflect.FileDescriptor

const file_github_com_moby_buildkit_session_auth_auth_proto_rawDesc = "" +
	"\n" +
	"0github.com/moby/buildkit/session/auth/auth.proto\x12\x10moby.filesync.v1\"(\n" +
	"\x12CredentialsRequest\x12\x12\n" +
	"\x04Host\x18\x01 \x01(\tR\x04Host\"I\n" +
	"\x13CredentialsResponse\x12\x1a\n" +
	"\bUsername\x18\x01 \x01(\tR\bUsername\x12\x16\n" +
	"\x06Secret\x18\x02 \x01(\tR\x06Secret\"\x8b\x01\n" +
	"\x11FetchTokenRequest\x12\x1a\n" +
	"\bClientID\x18\x01 \x01(\tR\bClientID\x12\x12\n" +
	"\x04Host\x18\x02 \x01(\tR\x04Host\x12\x14\n" +
	"\x05Realm\x18\x03 \x01(\tR\x05Realm\x12\x18\n" +
	"\aService\x18\x04 \x01(\tR\aService\x12\x16\n" +
	"\x06Scopes\x18\x05 \x03(\tR\x06Scopes\"d\n" +
	"\x12FetchTokenResponse\x12\x14\n" +
	"\x05Token\x18\x01 \x01(\tR\x05Token\x12\x1c\n" +
	"\tExpiresIn\x18\x02 \x01(\x03R\tExpiresIn\x12\x1a\n" +
	"\bIssuedAt\x18\x03 \x01(\x03R\bIssuedAt\"B\n" +
	"\x18GetTokenAuthorityRequest\x12\x12\n" +
	"\x04Host\x18\x01 \x01(\tR\x04Host\x12\x12\n" +
	"\x04Salt\x18\x02 \x01(\fR\x04Salt\"9\n" +
	"\x19GetTokenAuthorityResponse\x12\x1c\n" +
	"\tPublicKey\x18\x01 \x01(\fR\tPublicKey\"_\n" +
	"\x1bVerifyTokenAuthorityRequest\x12\x12\n" +
	"\x04Host\x18\x01 \x01(\tR\x04Host\x12\x18\n" +
	"\aPayload\x18\x02 \x01(\fR\aPayload\x12\x12\n" +
	"\x04Salt\x18\x03 \x01(\fR\x04Salt\"6\n" +
	"\x1cVerifyTokenAuthorityResponse\x12\x16\n" +
	"\x06Signed\x18\x01 \x01(\fR\x06Signed2\xa0\x03\n" +
	"\x04Auth\x12Z\n" +
	"\vCredentials\x12$.moby.filesync.v1.CredentialsRequest\x1a%.moby.filesync.v1.CredentialsResponse\x12W\n" +
	"\n" +
	"FetchToken\x12#.moby.filesync.v1.FetchTokenRequest\x1a$.moby.filesync.v1.FetchTokenResponse\x12l\n" +
	"\x11GetTokenAuthority\x12*.moby.filesync.v1.GetTokenAuthorityRequest\x1a+.moby.filesync.v1.GetTokenAuthorityResponse\x12u\n" +
	"\x14VerifyTokenAuthority\x12-.moby.filesync.v1.VerifyTokenAuthorityRequest\x1a..moby.filesync.v1.VerifyTokenAuthorityResponseB'Z%github.com/moby/buildkit/session/authb\x06proto3"

var (
	file_github_com_moby_buildkit_session_auth_auth_proto_rawDescOnce sync.Once
	file_github_com_moby_buildkit_session_auth_auth_proto_rawDescData []byte
)

func file_github_com_moby_buildkit_session_auth_auth_proto_rawDescGZIP() []byte {
	file_github_com_moby_buildkit_session_auth_auth_proto_rawDescOnce.Do(func() {
		file_github_com_moby_buildkit_session_auth_auth_proto_rawDescData = protoimpl.X.CompressGZIP(unsafe.Slice(unsafe.StringData(file_github_com_moby_buildkit_session_auth_auth_proto_rawDesc), len(file_github_com_moby_buildkit_session_auth_auth_proto_rawDesc)))
	})
	return file_github_com_moby_buildkit_session_auth_auth_proto_rawDescData
}

var file_github_com_moby_buildkit_session_auth_auth_proto_msgTypes = make([]protoimpl.MessageInfo, 8)
var file_github_com_moby_buildkit_session_auth_auth_proto_goTypes = []any{
	(*CredentialsRequest)(nil),           // 0: moby.filesync.v1.CredentialsRequest
	(*CredentialsResponse)(nil),          // 1: moby.filesync.v1.CredentialsResponse
	(*FetchTokenRequest)(nil),            // 2: moby.filesync.v1.FetchTokenRequest
	(*FetchTokenResponse)(nil),           // 3: moby.filesync.v1.FetchTokenResponse
	(*GetTokenAuthorityRequest)(nil),     // 4: moby.filesync.v1.GetTokenAuthorityRequest
	(*GetTokenAuthorityResponse)(nil),    // 5: moby.filesync.v1.GetTokenAuthorityResponse
	(*VerifyTokenAuthorityRequest)(nil),  // 6: moby.filesync.v1.VerifyTokenAuthorityRequest
	(*VerifyTokenAuthorityResponse)(nil), // 7: moby.filesync.v1.VerifyTokenAuthorityResponse
}
var file_github_com_moby_buildkit_session_auth_auth_proto_depIdxs = []int32{
	0, // 0: moby.filesync.v1.Auth.Credentials:input_type -> moby.filesync.v1.CredentialsRequest
	2, // 1: moby.filesync.v1.Auth.FetchToken:input_type -> moby.filesync.v1.FetchTokenRequest
	4, // 2: moby.filesync.v1.Auth.GetTokenAuthority:input_type -> moby.filesync.v1.GetTokenAuthorityRequest
	6, // 3: moby.filesync.v1.Auth.VerifyTokenAuthority:input_type -> moby.filesync.v1.VerifyTokenAuthorityRequest
	1, // 4: moby.filesync.v1.Auth.Credentials:output_type -> moby.filesync.v1.CredentialsResponse
	3, // 5: moby.filesync.v1.Auth.FetchToken:output_type -> moby.filesync.v1.FetchTokenResponse
	5, // 6: moby.filesync.v1.Auth.GetTokenAuthority:output_type -> moby.filesync.v1.GetTokenAuthorityResponse
	7, // 7: moby.filesync.v1.Auth.VerifyTokenAuthority:output_type -> moby.filesync.v1.VerifyTokenAuthorityResponse
	4, // [4:8] is the sub-list for method output_type
	0, // [0:4] is the sub-list for method input_type
	0, // [0:0] is the sub-list for extension type_name
	0, // [0:0] is the sub-list for extension extendee
	0, // [0:0] is the sub-list for field type_name
}

func init() { file_github_com_moby_buildkit_session_auth_auth_proto_init() }
func file_github_com_moby_buildkit_session_auth_auth_proto_init() {
	if File_github_com_moby_buildkit_session_auth_auth_proto != nil {
		return
	}
	type x struct{}
	out := protoimpl.TypeBuilder{
		File: protoimpl.DescBuilder{
			GoPackagePath: reflect.TypeOf(x{}).PkgPath(),
			RawDescriptor: unsafe.Slice(unsafe.StringData(file_github_com_moby_buildkit_session_auth_auth_proto_rawDesc), len(file_github_com_moby_buildkit_session_auth_auth_proto_rawDesc)),
			NumEnums:      0,
			NumMessages:   8,
			NumExtensions: 0,
			NumServices:   1,
		},
		GoTypes:           file_github_com_moby_buildkit_session_auth_auth_proto_goTypes,
		DependencyIndexes: file_github_com_moby_buildkit_session_auth_auth_proto_depIdxs,
		MessageInfos:      file_github_com_moby_buildkit_session_auth_auth_proto_msgTypes,
	}.Build()
	File_github_com_moby_buildkit_session_auth_auth_proto = out.File
	file_github_com_moby_buildkit_session_auth_auth_proto_goTypes = nil
	file_github_com_moby_buildkit_session_auth_auth_proto_depIdxs = nil
}
