//go:build !dfrunsecurity

package dockerfile2llb

import (
	"github.com/moby/buildkit/client/llb"
	"github.com/moby/buildkit/frontend/dockerfile/instructions"
)

func dispatchRunSecurity(_ *instructions.RunCommand) (llb.RunOption, error) {
	return nil, nil
}
