// Code generated by protoc-gen-go. DO NOT EDIT.
// versions:
// 	protoc-gen-go v1.36.9
// 	protoc        v3.11.4
// source: github.com/moby/buildkit/api/types/worker.proto

package moby_buildkit_v1_types

import (
	pb "github.com/moby/buildkit/solver/pb"
	protoreflect "google.golang.org/protobuf/reflect/protoreflect"
	protoimpl "google.golang.org/protobuf/runtime/protoimpl"
	reflect "reflect"
	sync "sync"
	unsafe "unsafe"
)

const (
	// Verify that this generated code is sufficiently up-to-date.
	_ = protoimpl.EnforceVersion(20 - protoimpl.MinVersion)
	// Verify that runtime/protoimpl is sufficiently up-to-date.
	_ = protoimpl.EnforceVersion(protoimpl.MaxVersion - 20)
)

type WorkerRecord struct {
	state           protoimpl.MessageState `protogen:"open.v1"`
	ID              string                 `protobuf:"bytes,1,opt,name=ID,proto3" json:"ID,omitempty"`
	Labels          map[string]string      `protobuf:"bytes,2,rep,name=Labels,proto3" json:"Labels,omitempty" protobuf_key:"bytes,1,opt,name=key" protobuf_val:"bytes,2,opt,name=value"`
	Platforms       []*pb.Platform         `protobuf:"bytes,3,rep,name=platforms,proto3" json:"platforms,omitempty"`
	GCPolicy        []*GCPolicy            `protobuf:"bytes,4,rep,name=GCPolicy,proto3" json:"GCPolicy,omitempty"`
	BuildkitVersion *BuildkitVersion       `protobuf:"bytes,5,opt,name=BuildkitVersion,proto3" json:"BuildkitVersion,omitempty"`
	CDIDevices      []*CDIDevice           `protobuf:"bytes,6,rep,name=CDIDevices,proto3" json:"CDIDevices,omitempty"`
	unknownFields   protoimpl.UnknownFields
	sizeCache       protoimpl.SizeCache
}

func (x *WorkerRecord) Reset() {
	*x = WorkerRecord{}
	mi := &file_github_com_moby_buildkit_api_types_worker_proto_msgTypes[0]
	ms := protoimpl.X.MessageStateOf(protoimpl.Pointer(x))
	ms.StoreMessageInfo(mi)
}

func (x *WorkerRecord) String() string {
	return protoimpl.X.MessageStringOf(x)
}

func (*WorkerRecord) ProtoMessage() {}

func (x *WorkerRecord) ProtoReflect() protoreflect.Message {
	mi := &file_github_com_moby_buildkit_api_types_worker_proto_msgTypes[0]
	if x != nil {
		ms := protoimpl.X.MessageStateOf(protoimpl.Pointer(x))
		if ms.LoadMessageInfo() == nil {
			ms.StoreMessageInfo(mi)
		}
		return ms
	}
	return mi.MessageOf(x)
}

// Deprecated: Use WorkerRecord.ProtoReflect.Descriptor instead.
func (*WorkerRecord) Descriptor() ([]byte, []int) {
	return file_github_com_moby_buildkit_api_types_worker_proto_rawDescGZIP(), []int{0}
}

func (x *WorkerRecord) GetID() string {
	if x != nil {
		return x.ID
	}
	return ""
}

func (x *WorkerRecord) GetLabels() map[string]string {
	if x != nil {
		return x.Labels
	}
	return nil
}

func (x *WorkerRecord) GetPlatforms() []*pb.Platform {
	if x != nil {
		return x.Platforms
	}
	return nil
}

func (x *WorkerRecord) GetGCPolicy() []*GCPolicy {
	if x != nil {
		return x.GCPolicy
	}
	return nil
}

func (x *WorkerRecord) GetBuildkitVersion() *BuildkitVersion {
	if x != nil {
		return x.BuildkitVersion
	}
	return nil
}

func (x *WorkerRecord) GetCDIDevices() []*CDIDevice {
	if x != nil {
		return x.CDIDevices
	}
	return nil
}

type GCPolicy struct {
	state        protoimpl.MessageState `protogen:"open.v1"`
	All          bool                   `protobuf:"varint,1,opt,name=all,proto3" json:"all,omitempty"`
	KeepDuration int64                  `protobuf:"varint,2,opt,name=keepDuration,proto3" json:"keepDuration,omitempty"`
	Filters      []string               `protobuf:"bytes,4,rep,name=filters,proto3" json:"filters,omitempty"`
	// reservedSpace was renamed from freeBytes
	ReservedSpace int64 `protobuf:"varint,3,opt,name=reservedSpace,proto3" json:"reservedSpace,omitempty"`
	MaxUsedSpace  int64 `protobuf:"varint,5,opt,name=maxUsedSpace,proto3" json:"maxUsedSpace,omitempty"`
	MinFreeSpace  int64 `protobuf:"varint,6,opt,name=minFreeSpace,proto3" json:"minFreeSpace,omitempty"`
	unknownFields protoimpl.UnknownFields
	sizeCache     protoimpl.SizeCache
}

func (x *GCPolicy) Reset() {
	*x = GCPolicy{}
	mi := &file_github_com_moby_buildkit_api_types_worker_proto_msgTypes[1]
	ms := protoimpl.X.MessageStateOf(protoimpl.Pointer(x))
	ms.StoreMessageInfo(mi)
}

func (x *GCPolicy) String() string {
	return protoimpl.X.MessageStringOf(x)
}

func (*GCPolicy) ProtoMessage() {}

func (x *GCPolicy) ProtoReflect() protoreflect.Message {
	mi := &file_github_com_moby_buildkit_api_types_worker_proto_msgTypes[1]
	if x != nil {
		ms := protoimpl.X.MessageStateOf(protoimpl.Pointer(x))
		if ms.LoadMessageInfo() == nil {
			ms.StoreMessageInfo(mi)
		}
		return ms
	}
	return mi.MessageOf(x)
}

// Deprecated: Use GCPolicy.ProtoReflect.Descriptor instead.
func (*GCPolicy) Descriptor() ([]byte, []int) {
	return file_github_com_moby_buildkit_api_types_worker_proto_rawDescGZIP(), []int{1}
}

func (x *GCPolicy) GetAll() bool {
	if x != nil {
		return x.All
	}
	return false
}

func (x *GCPolicy) GetKeepDuration() int64 {
	if x != nil {
		return x.KeepDuration
	}
	return 0
}

func (x *GCPolicy) GetFilters() []string {
	if x != nil {
		return x.Filters
	}
	return nil
}

func (x *GCPolicy) GetReservedSpace() int64 {
	if x != nil {
		return x.ReservedSpace
	}
	return 0
}

func (x *GCPolicy) GetMaxUsedSpace() int64 {
	if x != nil {
		return x.MaxUsedSpace
	}
	return 0
}

func (x *GCPolicy) GetMinFreeSpace() int64 {
	if x != nil {
		return x.MinFreeSpace
	}
	return 0
}

type BuildkitVersion struct {
	state         protoimpl.MessageState `protogen:"open.v1"`
	Package       string                 `protobuf:"bytes,1,opt,name=package,proto3" json:"package,omitempty"`
	Version       string                 `protobuf:"bytes,2,opt,name=version,proto3" json:"version,omitempty"`
	Revision      string                 `protobuf:"bytes,3,opt,name=revision,proto3" json:"revision,omitempty"`
	unknownFields protoimpl.UnknownFields
	sizeCache     protoimpl.SizeCache
}

func (x *BuildkitVersion) Reset() {
	*x = BuildkitVersion{}
	mi := &file_github_com_moby_buildkit_api_types_worker_proto_msgTypes[2]
	ms := protoimpl.X.MessageStateOf(protoimpl.Pointer(x))
	ms.StoreMessageInfo(mi)
}

func (x *BuildkitVersion) String() string {
	return protoimpl.X.MessageStringOf(x)
}

func (*BuildkitVersion) ProtoMessage() {}

func (x *BuildkitVersion) ProtoReflect() protoreflect.Message {
	mi := &file_github_com_moby_buildkit_api_types_worker_proto_msgTypes[2]
	if x != nil {
		ms := protoimpl.X.MessageStateOf(protoimpl.Pointer(x))
		if ms.LoadMessageInfo() == nil {
			ms.StoreMessageInfo(mi)
		}
		return ms
	}
	return mi.MessageOf(x)
}

// Deprecated: Use BuildkitVersion.ProtoReflect.Descriptor instead.
func (*BuildkitVersion) Descriptor() ([]byte, []int) {
	return file_github_com_moby_buildkit_api_types_worker_proto_rawDescGZIP(), []int{2}
}

func (x *BuildkitVersion) GetPackage() string {
	if x != nil {
		return x.Package
	}
	return ""
}

func (x *BuildkitVersion) GetVersion() string {
	if x != nil {
		return x.Version
	}
	return ""
}

func (x *BuildkitVersion) GetRevision() string {
	if x != nil {
		return x.Revision
	}
	return ""
}

type CDIDevice struct {
	state         protoimpl.MessageState `protogen:"open.v1"`
	Name          string                 `protobuf:"bytes,1,opt,name=Name,proto3" json:"Name,omitempty"`
	AutoAllow     bool                   `protobuf:"varint,2,opt,name=AutoAllow,proto3" json:"AutoAllow,omitempty"`
	Annotations   map[string]string      `protobuf:"bytes,3,rep,name=Annotations,proto3" json:"Annotations,omitempty" protobuf_key:"bytes,1,opt,name=key" protobuf_val:"bytes,2,opt,name=value"`
	OnDemand      bool                   `protobuf:"varint,4,opt,name=OnDemand,proto3" json:"OnDemand,omitempty"`
	unknownFields protoimpl.UnknownFields
	sizeCache     protoimpl.SizeCache
}

func (x *CDIDevice) Reset() {
	*x = CDIDevice{}
	mi := &file_github_com_moby_buildkit_api_types_worker_proto_msgTypes[3]
	ms := protoimpl.X.MessageStateOf(protoimpl.Pointer(x))
	ms.StoreMessageInfo(mi)
}

func (x *CDIDevice) String() string {
	return protoimpl.X.MessageStringOf(x)
}

func (*CDIDevice) ProtoMessage() {}

func (x *CDIDevice) ProtoReflect() protoreflect.Message {
	mi := &file_github_com_moby_buildkit_api_types_worker_proto_msgTypes[3]
	if x != nil {
		ms := protoimpl.X.MessageStateOf(protoimpl.Pointer(x))
		if ms.LoadMessageInfo() == nil {
			ms.StoreMessageInfo(mi)
		}
		return ms
	}
	return mi.MessageOf(x)
}

// Deprecated: Use CDIDevice.ProtoReflect.Descriptor instead.
func (*CDIDevice) Descriptor() ([]byte, []int) {
	return file_github_com_moby_buildkit_api_types_worker_proto_rawDescGZIP(), []int{3}
}

func (x *CDIDevice) GetName() string {
	if x != nil {
		return x.Name
	}
	return ""
}

func (x *CDIDevice) GetAutoAllow() bool {
	if x != nil {
		return x.AutoAllow
	}
	return false
}

func (x *CDIDevice) GetAnnotations() map[string]string {
	if x != nil {
		return x.Annotations
	}
	return nil
}

func (x *CDIDevice) GetOnDemand() bool {
	if x != nil {
		return x.OnDemand
	}
	return false
}

var File_github_com_moby_buildkit_api_types_worker_proto protoreflect.FileDescriptor

const file_github_com_moby_buildkit_api_types_worker_proto_rawDesc = "" +
	"\n" +
	"/github.com/moby/buildkit/api/types/worker.proto\x12\x16moby.buildkit.v1.types\x1a,github.com/moby/buildkit/solver/pb/ops.proto\"\xa3\x03\n" +
	"\fWorkerRecord\x12\x0e\n" +
	"\x02ID\x18\x01 \x01(\tR\x02ID\x12H\n" +
	"\x06Labels\x18\x02 \x03(\v20.moby.buildkit.v1.types.WorkerRecord.LabelsEntryR\x06Labels\x12*\n" +
	"\tplatforms\x18\x03 \x03(\v2\f.pb.PlatformR\tplatforms\x12<\n" +
	"\bGCPolicy\x18\x04 \x03(\v2 .moby.buildkit.v1.types.GCPolicyR\bGCPolicy\x12Q\n" +
	"\x0fBuildkitVersion\x18\x05 \x01(\v2'.moby.buildkit.v1.types.BuildkitVersionR\x0fBuildkitVersion\x12A\n" +
	"\n" +
	"CDIDevices\x18\x06 \x03(\v2!.moby.buildkit.v1.types.CDIDeviceR\n" +
	"CDIDevices\x1a9\n" +
	"\vLabelsEntry\x12\x10\n" +
	"\x03key\x18\x01 \x01(\tR\x03key\x12\x14\n" +
	"\x05value\x18\x02 \x01(\tR\x05value:\x028\x01\"\xc8\x01\n" +
	"\bGCPolicy\x12\x10\n" +
	"\x03all\x18\x01 \x01(\bR\x03all\x12\"\n" +
	"\fkeepDuration\x18\x02 \x01(\x03R\fkeepDuration\x12\x18\n" +
	"\afilters\x18\x04 \x03(\tR\afilters\x12$\n" +
	"\rreservedSpace\x18\x03 \x01(\x03R\rreservedSpace\x12\"\n" +
	"\fmaxUsedSpace\x18\x05 \x01(\x03R\fmaxUsedSpace\x12\"\n" +
	"\fminFreeSpace\x18\x06 \x01(\x03R\fminFreeSpace\"a\n" +
	"\x0fBuildkitVersion\x12\x18\n" +
	"\apackage\x18\x01 \x01(\tR\apackage\x12\x18\n" +
	"\aversion\x18\x02 \x01(\tR\aversion\x12\x1a\n" +
	"\brevision\x18\x03 \x01(\tR\brevision\"\xef\x01\n" +
	"\tCDIDevice\x12\x12\n" +
	"\x04Name\x18\x01 \x01(\tR\x04Name\x12\x1c\n" +
	"\tAutoAllow\x18\x02 \x01(\bR\tAutoAllow\x12T\n" +
	"\vAnnotations\x18\x03 \x03(\v22.moby.buildkit.v1.types.CDIDevice.AnnotationsEntryR\vAnnotations\x12\x1a\n" +
	"\bOnDemand\x18\x04 \x01(\bR\bOnDemand\x1a>\n" +
	"\x10AnnotationsEntry\x12\x10\n" +
	"\x03key\x18\x01 \x01(\tR\x03key\x12\x14\n" +
	"\x05value\x18\x02 \x01(\tR\x05value:\x028\x01B;Z9github.com/moby/buildkit/api/types;moby_buildkit_v1_typesb\x06proto3"

var (
	file_github_com_moby_buildkit_api_types_worker_proto_rawDescOnce sync.Once
	file_github_com_moby_buildkit_api_types_worker_proto_rawDescData []byte
)

func file_github_com_moby_buildkit_api_types_worker_proto_rawDescGZIP() []byte {
	file_github_com_moby_buildkit_api_types_worker_proto_rawDescOnce.Do(func() {
		file_github_com_moby_buildkit_api_types_worker_proto_rawDescData = protoimpl.X.CompressGZIP(unsafe.Slice(unsafe.StringData(file_github_com_moby_buildkit_api_types_worker_proto_rawDesc), len(file_github_com_moby_buildkit_api_types_worker_proto_rawDesc)))
	})
	return file_github_com_moby_buildkit_api_types_worker_proto_rawDescData
}

var file_github_com_moby_buildkit_api_types_worker_proto_msgTypes = make([]protoimpl.MessageInfo, 6)
var file_github_com_moby_buildkit_api_types_worker_proto_goTypes = []any{
	(*WorkerRecord)(nil),    // 0: moby.buildkit.v1.types.WorkerRecord
	(*GCPolicy)(nil),        // 1: moby.buildkit.v1.types.GCPolicy
	(*BuildkitVersion)(nil), // 2: moby.buildkit.v1.types.BuildkitVersion
	(*CDIDevice)(nil),       // 3: moby.buildkit.v1.types.CDIDevice
	nil,                     // 4: moby.buildkit.v1.types.WorkerRecord.LabelsEntry
	nil,                     // 5: moby.buildkit.v1.types.CDIDevice.AnnotationsEntry
	(*pb.Platform)(nil),     // 6: pb.Platform
}
var file_github_com_moby_buildkit_api_types_worker_proto_depIdxs = []int32{
	4, // 0: moby.buildkit.v1.types.WorkerRecord.Labels:type_name -> moby.buildkit.v1.types.WorkerRecord.LabelsEntry
	6, // 1: moby.buildkit.v1.types.WorkerRecord.platforms:type_name -> pb.Platform
	1, // 2: moby.buildkit.v1.types.WorkerRecord.GCPolicy:type_name -> moby.buildkit.v1.types.GCPolicy
	2, // 3: moby.buildkit.v1.types.WorkerRecord.BuildkitVersion:type_name -> moby.buildkit.v1.types.BuildkitVersion
	3, // 4: moby.buildkit.v1.types.WorkerRecord.CDIDevices:type_name -> moby.buildkit.v1.types.CDIDevice
	5, // 5: moby.buildkit.v1.types.CDIDevice.Annotations:type_name -> moby.buildkit.v1.types.CDIDevice.AnnotationsEntry
	6, // [6:6] is the sub-list for method output_type
	6, // [6:6] is the sub-list for method input_type
	6, // [6:6] is the sub-list for extension type_name
	6, // [6:6] is the sub-list for extension extendee
	0, // [0:6] is the sub-list for field type_name
}

func init() { file_github_com_moby_buildkit_api_types_worker_proto_init() }
func file_github_com_moby_buildkit_api_types_worker_proto_init() {
	if File_github_com_moby_buildkit_api_types_worker_proto != nil {
		return
	}
	type x struct{}
	out := protoimpl.TypeBuilder{
		File: protoimpl.DescBuilder{
			GoPackagePath: reflect.TypeOf(x{}).PkgPath(),
			RawDescriptor: unsafe.Slice(unsafe.StringData(file_github_com_moby_buildkit_api_types_worker_proto_rawDesc), len(file_github_com_moby_buildkit_api_types_worker_proto_rawDesc)),
			NumEnums:      0,
			NumMessages:   6,
			NumExtensions: 0,
			NumServices:   0,
		},
		GoTypes:           file_github_com_moby_buildkit_api_types_worker_proto_goTypes,
		DependencyIndexes: file_github_com_moby_buildkit_api_types_worker_proto_depIdxs,
		MessageInfos:      file_github_com_moby_buildkit_api_types_worker_proto_msgTypes,
	}.Build()
	File_github_com_moby_buildkit_api_types_worker_proto = out.File
	file_github_com_moby_buildkit_api_types_worker_proto_goTypes = nil
	file_github_com_moby_buildkit_api_types_worker_proto_depIdxs = nil
}
