// Code generated by protoc-gen-go. DO NOT EDIT.
// versions:
// 	protoc-gen-go v1.36.9
// 	protoc        v3.11.4
// source: github.com/moby/buildkit/cache/contenthash/checksum.proto

package contenthash

import (
	protoreflect "google.golang.org/protobuf/reflect/protoreflect"
	protoimpl "google.golang.org/protobuf/runtime/protoimpl"
	reflect "reflect"
	sync "sync"
	unsafe "unsafe"
)

const (
	// Verify that this generated code is sufficiently up-to-date.
	_ = protoimpl.EnforceVersion(20 - protoimpl.MinVersion)
	// Verify that runtime/protoimpl is sufficiently up-to-date.
	_ = protoimpl.EnforceVersion(protoimpl.MaxVersion - 20)
)

type CacheRecordType int32

const (
	CacheRecordType_FILE       CacheRecordType = 0
	CacheRecordType_DIR        CacheRecordType = 1
	CacheRecordType_DIR_HEADER CacheRecordType = 2
	CacheRecordType_SYMLINK    CacheRecordType = 3
)

// Enum value maps for CacheRecordType.
var (
	CacheRecordType_name = map[int32]string{
		0: "FILE",
		1: "DIR",
		2: "DIR_HEADER",
		3: "SYMLINK",
	}
	CacheRecordType_value = map[string]int32{
		"FILE":       0,
		"DIR":        1,
		"DIR_HEADER": 2,
		"SYMLINK":    3,
	}
)

func (x CacheRecordType) Enum() *CacheRecordType {
	p := new(CacheRecordType)
	*p = x
	return p
}

func (x CacheRecordType) String() string {
	return protoimpl.X.EnumStringOf(x.Descriptor(), protoreflect.EnumNumber(x))
}

func (CacheRecordType) Descriptor() protoreflect.EnumDescriptor {
	return file_github_com_moby_buildkit_cache_contenthash_checksum_proto_enumTypes[0].Descriptor()
}

func (CacheRecordType) Type() protoreflect.EnumType {
	return &file_github_com_moby_buildkit_cache_contenthash_checksum_proto_enumTypes[0]
}

func (x CacheRecordType) Number() protoreflect.EnumNumber {
	return protoreflect.EnumNumber(x)
}

// Deprecated: Use CacheRecordType.Descriptor instead.
func (CacheRecordType) EnumDescriptor() ([]byte, []int) {
	return file_github_com_moby_buildkit_cache_contenthash_checksum_proto_rawDescGZIP(), []int{0}
}

type CacheRecord struct {
	state         protoimpl.MessageState `protogen:"open.v1"`
	Digest        string                 `protobuf:"bytes,1,opt,name=digest,proto3" json:"digest,omitempty"`
	Type          CacheRecordType        `protobuf:"varint,2,opt,name=type,proto3,enum=contenthash.CacheRecordType" json:"type,omitempty"`
	Linkname      string                 `protobuf:"bytes,3,opt,name=linkname,proto3" json:"linkname,omitempty"`
	unknownFields protoimpl.UnknownFields
	sizeCache     protoimpl.SizeCache
}

func (x *CacheRecord) Reset() {
	*x = CacheRecord{}
	mi := &file_github_com_moby_buildkit_cache_contenthash_checksum_proto_msgTypes[0]
	ms := protoimpl.X.MessageStateOf(protoimpl.Pointer(x))
	ms.StoreMessageInfo(mi)
}

func (x *CacheRecord) String() string {
	return protoimpl.X.MessageStringOf(x)
}

func (*CacheRecord) ProtoMessage() {}

func (x *CacheRecord) ProtoReflect() protoreflect.Message {
	mi := &file_github_com_moby_buildkit_cache_contenthash_checksum_proto_msgTypes[0]
	if x != nil {
		ms := protoimpl.X.MessageStateOf(protoimpl.Pointer(x))
		if ms.LoadMessageInfo() == nil {
			ms.StoreMessageInfo(mi)
		}
		return ms
	}
	return mi.MessageOf(x)
}

// Deprecated: Use CacheRecord.ProtoReflect.Descriptor instead.
func (*CacheRecord) Descriptor() ([]byte, []int) {
	return file_github_com_moby_buildkit_cache_contenthash_checksum_proto_rawDescGZIP(), []int{0}
}

func (x *CacheRecord) GetDigest() string {
	if x != nil {
		return x.Digest
	}
	return ""
}

func (x *CacheRecord) GetType() CacheRecordType {
	if x != nil {
		return x.Type
	}
	return CacheRecordType_FILE
}

func (x *CacheRecord) GetLinkname() string {
	if x != nil {
		return x.Linkname
	}
	return ""
}

type CacheRecordWithPath struct {
	state         protoimpl.MessageState `protogen:"open.v1"`
	Path          string                 `protobuf:"bytes,1,opt,name=path,proto3" json:"path,omitempty"`
	Record        *CacheRecord           `protobuf:"bytes,2,opt,name=record,proto3" json:"record,omitempty"`
	unknownFields protoimpl.UnknownFields
	sizeCache     protoimpl.SizeCache
}

func (x *CacheRecordWithPath) Reset() {
	*x = CacheRecordWithPath{}
	mi := &file_github_com_moby_buildkit_cache_contenthash_checksum_proto_msgTypes[1]
	ms := protoimpl.X.MessageStateOf(protoimpl.Pointer(x))
	ms.StoreMessageInfo(mi)
}

func (x *CacheRecordWithPath) String() string {
	return protoimpl.X.MessageStringOf(x)
}

func (*CacheRecordWithPath) ProtoMessage() {}

func (x *CacheRecordWithPath) ProtoReflect() protoreflect.Message {
	mi := &file_github_com_moby_buildkit_cache_contenthash_checksum_proto_msgTypes[1]
	if x != nil {
		ms := protoimpl.X.MessageStateOf(protoimpl.Pointer(x))
		if ms.LoadMessageInfo() == nil {
			ms.StoreMessageInfo(mi)
		}
		return ms
	}
	return mi.MessageOf(x)
}

// Deprecated: Use CacheRecordWithPath.ProtoReflect.Descriptor instead.
func (*CacheRecordWithPath) Descriptor() ([]byte, []int) {
	return file_github_com_moby_buildkit_cache_contenthash_checksum_proto_rawDescGZIP(), []int{1}
}

func (x *CacheRecordWithPath) GetPath() string {
	if x != nil {
		return x.Path
	}
	return ""
}

func (x *CacheRecordWithPath) GetRecord() *CacheRecord {
	if x != nil {
		return x.Record
	}
	return nil
}

type CacheRecords struct {
	state         protoimpl.MessageState `protogen:"open.v1"`
	Paths         []*CacheRecordWithPath `protobuf:"bytes,1,rep,name=paths,proto3" json:"paths,omitempty"`
	unknownFields protoimpl.UnknownFields
	sizeCache     protoimpl.SizeCache
}

func (x *CacheRecords) Reset() {
	*x = CacheRecords{}
	mi := &file_github_com_moby_buildkit_cache_contenthash_checksum_proto_msgTypes[2]
	ms := protoimpl.X.MessageStateOf(protoimpl.Pointer(x))
	ms.StoreMessageInfo(mi)
}

func (x *CacheRecords) String() string {
	return protoimpl.X.MessageStringOf(x)
}

func (*CacheRecords) ProtoMessage() {}

func (x *CacheRecords) ProtoReflect() protoreflect.Message {
	mi := &file_github_com_moby_buildkit_cache_contenthash_checksum_proto_msgTypes[2]
	if x != nil {
		ms := protoimpl.X.MessageStateOf(protoimpl.Pointer(x))
		if ms.LoadMessageInfo() == nil {
			ms.StoreMessageInfo(mi)
		}
		return ms
	}
	return mi.MessageOf(x)
}

// Deprecated: Use CacheRecords.ProtoReflect.Descriptor instead.
func (*CacheRecords) Descriptor() ([]byte, []int) {
	return file_github_com_moby_buildkit_cache_contenthash_checksum_proto_rawDescGZIP(), []int{2}
}

func (x *CacheRecords) GetPaths() []*CacheRecordWithPath {
	if x != nil {
		return x.Paths
	}
	return nil
}

var File_github_com_moby_buildkit_cache_contenthash_checksum_proto protoreflect.FileDescriptor

const file_github_com_moby_buildkit_cache_contenthash_checksum_proto_rawDesc = "" +
	"\n" +
	"9github.com/moby/buildkit/cache/contenthash/checksum.proto\x12\vcontenthash\"s\n" +
	"\vCacheRecord\x12\x16\n" +
	"\x06digest\x18\x01 \x01(\tR\x06digest\x120\n" +
	"\x04type\x18\x02 \x01(\x0e2\x1c.contenthash.CacheRecordTypeR\x04type\x12\x1a\n" +
	"\blinkname\x18\x03 \x01(\tR\blinkname\"[\n" +
	"\x13CacheRecordWithPath\x12\x12\n" +
	"\x04path\x18\x01 \x01(\tR\x04path\x120\n" +
	"\x06record\x18\x02 \x01(\v2\x18.contenthash.CacheRecordR\x06record\"F\n" +
	"\fCacheRecords\x126\n" +
	"\x05paths\x18\x01 \x03(\v2 .contenthash.CacheRecordWithPathR\x05paths*A\n" +
	"\x0fCacheRecordType\x12\b\n" +
	"\x04FILE\x10\x00\x12\a\n" +
	"\x03DIR\x10\x01\x12\x0e\n" +
	"\n" +
	"DIR_HEADER\x10\x02\x12\v\n" +
	"\aSYMLINK\x10\x03B,Z*github.com/moby/buildkit/cache/contenthashb\x06proto3"

var (
	file_github_com_moby_buildkit_cache_contenthash_checksum_proto_rawDescOnce sync.Once
	file_github_com_moby_buildkit_cache_contenthash_checksum_proto_rawDescData []byte
)

func file_github_com_moby_buildkit_cache_contenthash_checksum_proto_rawDescGZIP() []byte {
	file_github_com_moby_buildkit_cache_contenthash_checksum_proto_rawDescOnce.Do(func() {
		file_github_com_moby_buildkit_cache_contenthash_checksum_proto_rawDescData = protoimpl.X.CompressGZIP(unsafe.Slice(unsafe.StringData(file_github_com_moby_buildkit_cache_contenthash_checksum_proto_rawDesc), len(file_github_com_moby_buildkit_cache_contenthash_checksum_proto_rawDesc)))
	})
	return file_github_com_moby_buildkit_cache_contenthash_checksum_proto_rawDescData
}

var file_github_com_moby_buildkit_cache_contenthash_checksum_proto_enumTypes = make([]protoimpl.EnumInfo, 1)
var file_github_com_moby_buildkit_cache_contenthash_checksum_proto_msgTypes = make([]protoimpl.MessageInfo, 3)
var file_github_com_moby_buildkit_cache_contenthash_checksum_proto_goTypes = []any{
	(CacheRecordType)(0),        // 0: contenthash.CacheRecordType
	(*CacheRecord)(nil),         // 1: contenthash.CacheRecord
	(*CacheRecordWithPath)(nil), // 2: contenthash.CacheRecordWithPath
	(*CacheRecords)(nil),        // 3: contenthash.CacheRecords
}
var file_github_com_moby_buildkit_cache_contenthash_checksum_proto_depIdxs = []int32{
	0, // 0: contenthash.CacheRecord.type:type_name -> contenthash.CacheRecordType
	1, // 1: contenthash.CacheRecordWithPath.record:type_name -> contenthash.CacheRecord
	2, // 2: contenthash.CacheRecords.paths:type_name -> contenthash.CacheRecordWithPath
	3, // [3:3] is the sub-list for method output_type
	3, // [3:3] is the sub-list for method input_type
	3, // [3:3] is the sub-list for extension type_name
	3, // [3:3] is the sub-list for extension extendee
	0, // [0:3] is the sub-list for field type_name
}

func init() { file_github_com_moby_buildkit_cache_contenthash_checksum_proto_init() }
func file_github_com_moby_buildkit_cache_contenthash_checksum_proto_init() {
	if File_github_com_moby_buildkit_cache_contenthash_checksum_proto != nil {
		return
	}
	type x struct{}
	out := protoimpl.TypeBuilder{
		File: protoimpl.DescBuilder{
			GoPackagePath: reflect.TypeOf(x{}).PkgPath(),
			RawDescriptor: unsafe.Slice(unsafe.StringData(file_github_com_moby_buildkit_cache_contenthash_checksum_proto_rawDesc), len(file_github_com_moby_buildkit_cache_contenthash_checksum_proto_rawDesc)),
			NumEnums:      1,
			NumMessages:   3,
			NumExtensions: 0,
			NumServices:   0,
		},
		GoTypes:           file_github_com_moby_buildkit_cache_contenthash_checksum_proto_goTypes,
		DependencyIndexes: file_github_com_moby_buildkit_cache_contenthash_checksum_proto_depIdxs,
		EnumInfos:         file_github_com_moby_buildkit_cache_contenthash_checksum_proto_enumTypes,
		MessageInfos:      file_github_com_moby_buildkit_cache_contenthash_checksum_proto_msgTypes,
	}.Build()
	File_github_com_moby_buildkit_cache_contenthash_checksum_proto = out.File
	file_github_com_moby_buildkit_cache_contenthash_checksum_proto_goTypes = nil
	file_github_com_moby_buildkit_cache_contenthash_checksum_proto_depIdxs = nil
}
