// Code generated by protoc-gen-go. DO NOT EDIT.
// versions:
// 	protoc-gen-go v1.36.9
// 	protoc        v3.11.4
// source: github.com/moby/buildkit/session/sshforward/ssh.proto

package sshforward

import (
	protoreflect "google.golang.org/protobuf/reflect/protoreflect"
	protoimpl "google.golang.org/protobuf/runtime/protoimpl"
	reflect "reflect"
	sync "sync"
	unsafe "unsafe"
)

const (
	// Verify that this generated code is sufficiently up-to-date.
	_ = protoimpl.EnforceVersion(20 - protoimpl.MinVersion)
	// Verify that runtime/protoimpl is sufficiently up-to-date.
	_ = protoimpl.EnforceVersion(protoimpl.MaxVersion - 20)
)

// BytesMessage contains a chunk of byte data
type BytesMessage struct {
	state         protoimpl.MessageState `protogen:"open.v1"`
	Data          []byte                 `protobuf:"bytes,1,opt,name=data,proto3" json:"data,omitempty"`
	unknownFields protoimpl.UnknownFields
	sizeCache     protoimpl.SizeCache
}

func (x *BytesMessage) Reset() {
	*x = BytesMessage{}
	mi := &file_github_com_moby_buildkit_session_sshforward_ssh_proto_msgTypes[0]
	ms := protoimpl.X.MessageStateOf(protoimpl.Pointer(x))
	ms.StoreMessageInfo(mi)
}

func (x *BytesMessage) String() string {
	return protoimpl.X.MessageStringOf(x)
}

func (*BytesMessage) ProtoMessage() {}

func (x *BytesMessage) ProtoReflect() protoreflect.Message {
	mi := &file_github_com_moby_buildkit_session_sshforward_ssh_proto_msgTypes[0]
	if x != nil {
		ms := protoimpl.X.MessageStateOf(protoimpl.Pointer(x))
		if ms.LoadMessageInfo() == nil {
			ms.StoreMessageInfo(mi)
		}
		return ms
	}
	return mi.MessageOf(x)
}

// Deprecated: Use BytesMessage.ProtoReflect.Descriptor instead.
func (*BytesMessage) Descriptor() ([]byte, []int) {
	return file_github_com_moby_buildkit_session_sshforward_ssh_proto_rawDescGZIP(), []int{0}
}

func (x *BytesMessage) GetData() []byte {
	if x != nil {
		return x.Data
	}
	return nil
}

type CheckAgentRequest struct {
	state         protoimpl.MessageState `protogen:"open.v1"`
	ID            string                 `protobuf:"bytes,1,opt,name=ID,proto3" json:"ID,omitempty"`
	unknownFields protoimpl.UnknownFields
	sizeCache     protoimpl.SizeCache
}

func (x *CheckAgentRequest) Reset() {
	*x = CheckAgentRequest{}
	mi := &file_github_com_moby_buildkit_session_sshforward_ssh_proto_msgTypes[1]
	ms := protoimpl.X.MessageStateOf(protoimpl.Pointer(x))
	ms.StoreMessageInfo(mi)
}

func (x *CheckAgentRequest) String() string {
	return protoimpl.X.MessageStringOf(x)
}

func (*CheckAgentRequest) ProtoMessage() {}

func (x *CheckAgentRequest) ProtoReflect() protoreflect.Message {
	mi := &file_github_com_moby_buildkit_session_sshforward_ssh_proto_msgTypes[1]
	if x != nil {
		ms := protoimpl.X.MessageStateOf(protoimpl.Pointer(x))
		if ms.LoadMessageInfo() == nil {
			ms.StoreMessageInfo(mi)
		}
		return ms
	}
	return mi.MessageOf(x)
}

// Deprecated: Use CheckAgentRequest.ProtoReflect.Descriptor instead.
func (*CheckAgentRequest) Descriptor() ([]byte, []int) {
	return file_github_com_moby_buildkit_session_sshforward_ssh_proto_rawDescGZIP(), []int{1}
}

func (x *CheckAgentRequest) GetID() string {
	if x != nil {
		return x.ID
	}
	return ""
}

type CheckAgentResponse struct {
	state         protoimpl.MessageState `protogen:"open.v1"`
	unknownFields protoimpl.UnknownFields
	sizeCache     protoimpl.SizeCache
}

func (x *CheckAgentResponse) Reset() {
	*x = CheckAgentResponse{}
	mi := &file_github_com_moby_buildkit_session_sshforward_ssh_proto_msgTypes[2]
	ms := protoimpl.X.MessageStateOf(protoimpl.Pointer(x))
	ms.StoreMessageInfo(mi)
}

func (x *CheckAgentResponse) String() string {
	return protoimpl.X.MessageStringOf(x)
}

func (*CheckAgentResponse) ProtoMessage() {}

func (x *CheckAgentResponse) ProtoReflect() protoreflect.Message {
	mi := &file_github_com_moby_buildkit_session_sshforward_ssh_proto_msgTypes[2]
	if x != nil {
		ms := protoimpl.X.MessageStateOf(protoimpl.Pointer(x))
		if ms.LoadMessageInfo() == nil {
			ms.StoreMessageInfo(mi)
		}
		return ms
	}
	return mi.MessageOf(x)
}

// Deprecated: Use CheckAgentResponse.ProtoReflect.Descriptor instead.
func (*CheckAgentResponse) Descriptor() ([]byte, []int) {
	return file_github_com_moby_buildkit_session_sshforward_ssh_proto_rawDescGZIP(), []int{2}
}

var File_github_com_moby_buildkit_session_sshforward_ssh_proto protoreflect.FileDescriptor

const file_github_com_moby_buildkit_session_sshforward_ssh_proto_rawDesc = "" +
	"\n" +
	"5github.com/moby/buildkit/session/sshforward/ssh.proto\x12\x12moby.sshforward.v1\"\"\n" +
	"\fBytesMessage\x12\x12\n" +
	"\x04data\x18\x01 \x01(\fR\x04data\"#\n" +
	"\x11CheckAgentRequest\x12\x0e\n" +
	"\x02ID\x18\x01 \x01(\tR\x02ID\"\x14\n" +
	"\x12CheckAgentResponse2\xba\x01\n" +
	"\x03SSH\x12[\n" +
	"\n" +
	"CheckAgent\x12%.moby.sshforward.v1.CheckAgentRequest\x1a&.moby.sshforward.v1.CheckAgentResponse\x12V\n" +
	"\fForwardAgent\x12 .moby.sshforward.v1.BytesMessage\x1a .moby.sshforward.v1.BytesMessage(\x010\x01B-Z+github.com/moby/buildkit/session/sshforwardb\x06proto3"

var (
	file_github_com_moby_buildkit_session_sshforward_ssh_proto_rawDescOnce sync.Once
	file_github_com_moby_buildkit_session_sshforward_ssh_proto_rawDescData []byte
)

func file_github_com_moby_buildkit_session_sshforward_ssh_proto_rawDescGZIP() []byte {
	file_github_com_moby_buildkit_session_sshforward_ssh_proto_rawDescOnce.Do(func() {
		file_github_com_moby_buildkit_session_sshforward_ssh_proto_rawDescData = protoimpl.X.CompressGZIP(unsafe.Slice(unsafe.StringData(file_github_com_moby_buildkit_session_sshforward_ssh_proto_rawDesc), len(file_github_com_moby_buildkit_session_sshforward_ssh_proto_rawDesc)))
	})
	return file_github_com_moby_buildkit_session_sshforward_ssh_proto_rawDescData
}

var file_github_com_moby_buildkit_session_sshforward_ssh_proto_msgTypes = make([]protoimpl.MessageInfo, 3)
var file_github_com_moby_buildkit_session_sshforward_ssh_proto_goTypes = []any{
	(*BytesMessage)(nil),       // 0: moby.sshforward.v1.BytesMessage
	(*CheckAgentRequest)(nil),  // 1: moby.sshforward.v1.CheckAgentRequest
	(*CheckAgentResponse)(nil), // 2: moby.sshforward.v1.CheckAgentResponse
}
var file_github_com_moby_buildkit_session_sshforward_ssh_proto_depIdxs = []int32{
	1, // 0: moby.sshforward.v1.SSH.CheckAgent:input_type -> moby.sshforward.v1.CheckAgentRequest
	0, // 1: moby.sshforward.v1.SSH.ForwardAgent:input_type -> moby.sshforward.v1.BytesMessage
	2, // 2: moby.sshforward.v1.SSH.CheckAgent:output_type -> moby.sshforward.v1.CheckAgentResponse
	0, // 3: moby.sshforward.v1.SSH.ForwardAgent:output_type -> moby.sshforward.v1.BytesMessage
	2, // [2:4] is the sub-list for method output_type
	0, // [0:2] is the sub-list for method input_type
	0, // [0:0] is the sub-list for extension type_name
	0, // [0:0] is the sub-list for extension extendee
	0, // [0:0] is the sub-list for field type_name
}

func init() { file_github_com_moby_buildkit_session_sshforward_ssh_proto_init() }
func file_github_com_moby_buildkit_session_sshforward_ssh_proto_init() {
	if File_github_com_moby_buildkit_session_sshforward_ssh_proto != nil {
		return
	}
	type x struct{}
	out := protoimpl.TypeBuilder{
		File: protoimpl.DescBuilder{
			GoPackagePath: reflect.TypeOf(x{}).PkgPath(),
			RawDescriptor: unsafe.Slice(unsafe.StringData(file_github_com_moby_buildkit_session_sshforward_ssh_proto_rawDesc), len(file_github_com_moby_buildkit_session_sshforward_ssh_proto_rawDesc)),
			NumEnums:      0,
			NumMessages:   3,
			NumExtensions: 0,
			NumServices:   1,
		},
		GoTypes:           file_github_com_moby_buildkit_session_sshforward_ssh_proto_goTypes,
		DependencyIndexes: file_github_com_moby_buildkit_session_sshforward_ssh_proto_depIdxs,
		MessageInfos:      file_github_com_moby_buildkit_session_sshforward_ssh_proto_msgTypes,
	}.Build()
	File_github_com_moby_buildkit_session_sshforward_ssh_proto = out.File
	file_github_com_moby_buildkit_session_sshforward_ssh_proto_goTypes = nil
	file_github_com_moby_buildkit_session_sshforward_ssh_proto_depIdxs = nil
}
