package attestation

const (
	DockerAnnotationReferenceType        = "vnd.docker.reference.type"
	DockerAnnotationReferenceDigest      = "vnd.docker.reference.digest"
	DockerAnnotationReferenceDescription = "vnd.docker.reference.description"

	DockerAnnotationReferenceTypeDefault = "attestation-manifest"
)
