// Code generated by protoc-gen-go. DO NOT EDIT.
// versions:
// 	protoc-gen-go v1.36.9
// 	protoc        v3.11.4
// source: github.com/moby/buildkit/api/services/control/control.proto

package moby_buildkit_v1

import (
	timestamp "github.com/golang/protobuf/ptypes/timestamp"
	types "github.com/moby/buildkit/api/types"
	pb "github.com/moby/buildkit/solver/pb"
	pb1 "github.com/moby/buildkit/sourcepolicy/pb"
	status "google.golang.org/genproto/googleapis/rpc/status"
	protoreflect "google.golang.org/protobuf/reflect/protoreflect"
	protoimpl "google.golang.org/protobuf/runtime/protoimpl"
	reflect "reflect"
	sync "sync"
	unsafe "unsafe"
)

const (
	// Verify that this generated code is sufficiently up-to-date.
	_ = protoimpl.EnforceVersion(20 - protoimpl.MinVersion)
	// Verify that runtime/protoimpl is sufficiently up-to-date.
	_ = protoimpl.EnforceVersion(protoimpl.MaxVersion - 20)
)

type BuildHistoryEventType int32

const (
	BuildHistoryEventType_STARTED  BuildHistoryEventType = 0
	BuildHistoryEventType_COMPLETE BuildHistoryEventType = 1
	BuildHistoryEventType_DELETED  BuildHistoryEventType = 2
)

// Enum value maps for BuildHistoryEventType.
var (
	BuildHistoryEventType_name = map[int32]string{
		0: "STARTED",
		1: "COMPLETE",
		2: "DELETED",
	}
	BuildHistoryEventType_value = map[string]int32{
		"STARTED":  0,
		"COMPLETE": 1,
		"DELETED":  2,
	}
)

func (x BuildHistoryEventType) Enum() *BuildHistoryEventType {
	p := new(BuildHistoryEventType)
	*p = x
	return p
}

func (x BuildHistoryEventType) String() string {
	return protoimpl.X.EnumStringOf(x.Descriptor(), protoreflect.EnumNumber(x))
}

func (BuildHistoryEventType) Descriptor() protoreflect.EnumDescriptor {
	return file_github_com_moby_buildkit_api_services_control_control_proto_enumTypes[0].Descriptor()
}

func (BuildHistoryEventType) Type() protoreflect.EnumType {
	return &file_github_com_moby_buildkit_api_services_control_control_proto_enumTypes[0]
}

func (x BuildHistoryEventType) Number() protoreflect.EnumNumber {
	return protoreflect.EnumNumber(x)
}

// Deprecated: Use BuildHistoryEventType.Descriptor instead.
func (BuildHistoryEventType) EnumDescriptor() ([]byte, []int) {
	return file_github_com_moby_buildkit_api_services_control_control_proto_rawDescGZIP(), []int{0}
}

type PruneRequest struct {
	state         protoimpl.MessageState `protogen:"open.v1"`
	Filter        []string               `protobuf:"bytes,1,rep,name=filter,proto3" json:"filter,omitempty"`
	All           bool                   `protobuf:"varint,2,opt,name=all,proto3" json:"all,omitempty"`
	KeepDuration  int64                  `protobuf:"varint,3,opt,name=keepDuration,proto3" json:"keepDuration,omitempty"`
	ReservedSpace int64                  `protobuf:"varint,4,opt,name=reservedSpace,proto3" json:"reservedSpace,omitempty"`
	MaxUsedSpace  int64                  `protobuf:"varint,5,opt,name=maxUsedSpace,proto3" json:"maxUsedSpace,omitempty"`
	MinFreeSpace  int64                  `protobuf:"varint,6,opt,name=minFreeSpace,proto3" json:"minFreeSpace,omitempty"`
	unknownFields protoimpl.UnknownFields
	sizeCache     protoimpl.SizeCache
}

func (x *PruneRequest) Reset() {
	*x = PruneRequest{}
	mi := &file_github_com_moby_buildkit_api_services_control_control_proto_msgTypes[0]
	ms := protoimpl.X.MessageStateOf(protoimpl.Pointer(x))
	ms.StoreMessageInfo(mi)
}

func (x *PruneRequest) String() string {
	return protoimpl.X.MessageStringOf(x)
}

func (*PruneRequest) ProtoMessage() {}

func (x *PruneRequest) ProtoReflect() protoreflect.Message {
	mi := &file_github_com_moby_buildkit_api_services_control_control_proto_msgTypes[0]
	if x != nil {
		ms := protoimpl.X.MessageStateOf(protoimpl.Pointer(x))
		if ms.LoadMessageInfo() == nil {
			ms.StoreMessageInfo(mi)
		}
		return ms
	}
	return mi.MessageOf(x)
}

// Deprecated: Use PruneRequest.ProtoReflect.Descriptor instead.
func (*PruneRequest) Descriptor() ([]byte, []int) {
	return file_github_com_moby_buildkit_api_services_control_control_proto_rawDescGZIP(), []int{0}
}

func (x *PruneRequest) GetFilter() []string {
	if x != nil {
		return x.Filter
	}
	return nil
}

func (x *PruneRequest) GetAll() bool {
	if x != nil {
		return x.All
	}
	return false
}

func (x *PruneRequest) GetKeepDuration() int64 {
	if x != nil {
		return x.KeepDuration
	}
	return 0
}

func (x *PruneRequest) GetReservedSpace() int64 {
	if x != nil {
		return x.ReservedSpace
	}
	return 0
}

func (x *PruneRequest) GetMaxUsedSpace() int64 {
	if x != nil {
		return x.MaxUsedSpace
	}
	return 0
}

func (x *PruneRequest) GetMinFreeSpace() int64 {
	if x != nil {
		return x.MinFreeSpace
	}
	return 0
}

type DiskUsageRequest struct {
	state         protoimpl.MessageState `protogen:"open.v1"`
	Filter        []string               `protobuf:"bytes,1,rep,name=filter,proto3" json:"filter,omitempty"`
	AgeLimit      int64                  `protobuf:"varint,2,opt,name=ageLimit,proto3" json:"ageLimit,omitempty"`
	unknownFields protoimpl.UnknownFields
	sizeCache     protoimpl.SizeCache
}

func (x *DiskUsageRequest) Reset() {
	*x = DiskUsageRequest{}
	mi := &file_github_com_moby_buildkit_api_services_control_control_proto_msgTypes[1]
	ms := protoimpl.X.MessageStateOf(protoimpl.Pointer(x))
	ms.StoreMessageInfo(mi)
}

func (x *DiskUsageRequest) String() string {
	return protoimpl.X.MessageStringOf(x)
}

func (*DiskUsageRequest) ProtoMessage() {}

func (x *DiskUsageRequest) ProtoReflect() protoreflect.Message {
	mi := &file_github_com_moby_buildkit_api_services_control_control_proto_msgTypes[1]
	if x != nil {
		ms := protoimpl.X.MessageStateOf(protoimpl.Pointer(x))
		if ms.LoadMessageInfo() == nil {
			ms.StoreMessageInfo(mi)
		}
		return ms
	}
	return mi.MessageOf(x)
}

// Deprecated: Use DiskUsageRequest.ProtoReflect.Descriptor instead.
func (*DiskUsageRequest) Descriptor() ([]byte, []int) {
	return file_github_com_moby_buildkit_api_services_control_control_proto_rawDescGZIP(), []int{1}
}

func (x *DiskUsageRequest) GetFilter() []string {
	if x != nil {
		return x.Filter
	}
	return nil
}

func (x *DiskUsageRequest) GetAgeLimit() int64 {
	if x != nil {
		return x.AgeLimit
	}
	return 0
}

type DiskUsageResponse struct {
	state         protoimpl.MessageState `protogen:"open.v1"`
	Record        []*UsageRecord         `protobuf:"bytes,1,rep,name=record,proto3" json:"record,omitempty"`
	unknownFields protoimpl.UnknownFields
	sizeCache     protoimpl.SizeCache
}

func (x *DiskUsageResponse) Reset() {
	*x = DiskUsageResponse{}
	mi := &file_github_com_moby_buildkit_api_services_control_control_proto_msgTypes[2]
	ms := protoimpl.X.MessageStateOf(protoimpl.Pointer(x))
	ms.StoreMessageInfo(mi)
}

func (x *DiskUsageResponse) String() string {
	return protoimpl.X.MessageStringOf(x)
}

func (*DiskUsageResponse) ProtoMessage() {}

func (x *DiskUsageResponse) ProtoReflect() protoreflect.Message {
	mi := &file_github_com_moby_buildkit_api_services_control_control_proto_msgTypes[2]
	if x != nil {
		ms := protoimpl.X.MessageStateOf(protoimpl.Pointer(x))
		if ms.LoadMessageInfo() == nil {
			ms.StoreMessageInfo(mi)
		}
		return ms
	}
	return mi.MessageOf(x)
}

// Deprecated: Use DiskUsageResponse.ProtoReflect.Descriptor instead.
func (*DiskUsageResponse) Descriptor() ([]byte, []int) {
	return file_github_com_moby_buildkit_api_services_control_control_proto_rawDescGZIP(), []int{2}
}

func (x *DiskUsageResponse) GetRecord() []*UsageRecord {
	if x != nil {
		return x.Record
	}
	return nil
}

type UsageRecord struct {
	state   protoimpl.MessageState `protogen:"open.v1"`
	ID      string                 `protobuf:"bytes,1,opt,name=ID,proto3" json:"ID,omitempty"`
	Mutable bool                   `protobuf:"varint,2,opt,name=Mutable,proto3" json:"Mutable,omitempty"`
	InUse   bool                   `protobuf:"varint,3,opt,name=InUse,proto3" json:"InUse,omitempty"`
	Size    int64                  `protobuf:"varint,4,opt,name=Size,proto3" json:"Size,omitempty"`
	// Deprecated: Marked as deprecated in github.com/moby/buildkit/api/services/control/control.proto.
	Parent        string               `protobuf:"bytes,5,opt,name=Parent,proto3" json:"Parent,omitempty"`
	CreatedAt     *timestamp.Timestamp `protobuf:"bytes,6,opt,name=CreatedAt,proto3" json:"CreatedAt,omitempty"`
	LastUsedAt    *timestamp.Timestamp `protobuf:"bytes,7,opt,name=LastUsedAt,proto3" json:"LastUsedAt,omitempty"`
	UsageCount    int64                `protobuf:"varint,8,opt,name=UsageCount,proto3" json:"UsageCount,omitempty"`
	Description   string               `protobuf:"bytes,9,opt,name=Description,proto3" json:"Description,omitempty"`
	RecordType    string               `protobuf:"bytes,10,opt,name=RecordType,proto3" json:"RecordType,omitempty"`
	Shared        bool                 `protobuf:"varint,11,opt,name=Shared,proto3" json:"Shared,omitempty"`
	Parents       []string             `protobuf:"bytes,12,rep,name=Parents,proto3" json:"Parents,omitempty"`
	unknownFields protoimpl.UnknownFields
	sizeCache     protoimpl.SizeCache
}

func (x *UsageRecord) Reset() {
	*x = UsageRecord{}
	mi := &file_github_com_moby_buildkit_api_services_control_control_proto_msgTypes[3]
	ms := protoimpl.X.MessageStateOf(protoimpl.Pointer(x))
	ms.StoreMessageInfo(mi)
}

func (x *UsageRecord) String() string {
	return protoimpl.X.MessageStringOf(x)
}

func (*UsageRecord) ProtoMessage() {}

func (x *UsageRecord) ProtoReflect() protoreflect.Message {
	mi := &file_github_com_moby_buildkit_api_services_control_control_proto_msgTypes[3]
	if x != nil {
		ms := protoimpl.X.MessageStateOf(protoimpl.Pointer(x))
		if ms.LoadMessageInfo() == nil {
			ms.StoreMessageInfo(mi)
		}
		return ms
	}
	return mi.MessageOf(x)
}

// Deprecated: Use UsageRecord.ProtoReflect.Descriptor instead.
func (*UsageRecord) Descriptor() ([]byte, []int) {
	return file_github_com_moby_buildkit_api_services_control_control_proto_rawDescGZIP(), []int{3}
}

func (x *UsageRecord) GetID() string {
	if x != nil {
		return x.ID
	}
	return ""
}

func (x *UsageRecord) GetMutable() bool {
	if x != nil {
		return x.Mutable
	}
	return false
}

func (x *UsageRecord) GetInUse() bool {
	if x != nil {
		return x.InUse
	}
	return false
}

func (x *UsageRecord) GetSize() int64 {
	if x != nil {
		return x.Size
	}
	return 0
}

// Deprecated: Marked as deprecated in github.com/moby/buildkit/api/services/control/control.proto.
func (x *UsageRecord) GetParent() string {
	if x != nil {
		return x.Parent
	}
	return ""
}

func (x *UsageRecord) GetCreatedAt() *timestamp.Timestamp {
	if x != nil {
		return x.CreatedAt
	}
	return nil
}

func (x *UsageRecord) GetLastUsedAt() *timestamp.Timestamp {
	if x != nil {
		return x.LastUsedAt
	}
	return nil
}

func (x *UsageRecord) GetUsageCount() int64 {
	if x != nil {
		return x.UsageCount
	}
	return 0
}

func (x *UsageRecord) GetDescription() string {
	if x != nil {
		return x.Description
	}
	return ""
}

func (x *UsageRecord) GetRecordType() string {
	if x != nil {
		return x.RecordType
	}
	return ""
}

func (x *UsageRecord) GetShared() bool {
	if x != nil {
		return x.Shared
	}
	return false
}

func (x *UsageRecord) GetParents() []string {
	if x != nil {
		return x.Parents
	}
	return nil
}

type SolveRequest struct {
	state      protoimpl.MessageState `protogen:"open.v1"`
	Ref        string                 `protobuf:"bytes,1,opt,name=Ref,proto3" json:"Ref,omitempty"`
	Definition *pb.Definition         `protobuf:"bytes,2,opt,name=Definition,proto3" json:"Definition,omitempty"`
	// ExporterDeprecated and ExporterAttrsDeprecated are deprecated in favor
	// of the new Exporters. If these fields are set, then they will be
	// appended to the Exporters field if Exporters was not explicitly set.
	ExporterDeprecated      string                    `protobuf:"bytes,3,opt,name=ExporterDeprecated,proto3" json:"ExporterDeprecated,omitempty"`
	ExporterAttrsDeprecated map[string]string         `protobuf:"bytes,4,rep,name=ExporterAttrsDeprecated,proto3" json:"ExporterAttrsDeprecated,omitempty" protobuf_key:"bytes,1,opt,name=key" protobuf_val:"bytes,2,opt,name=value"`
	Session                 string                    `protobuf:"bytes,5,opt,name=Session,proto3" json:"Session,omitempty"`
	Frontend                string                    `protobuf:"bytes,6,opt,name=Frontend,proto3" json:"Frontend,omitempty"`
	FrontendAttrs           map[string]string         `protobuf:"bytes,7,rep,name=FrontendAttrs,proto3" json:"FrontendAttrs,omitempty" protobuf_key:"bytes,1,opt,name=key" protobuf_val:"bytes,2,opt,name=value"`
	Cache                   *CacheOptions             `protobuf:"bytes,8,opt,name=Cache,proto3" json:"Cache,omitempty"`
	Entitlements            []string                  `protobuf:"bytes,9,rep,name=Entitlements,proto3" json:"Entitlements,omitempty"`
	FrontendInputs          map[string]*pb.Definition `protobuf:"bytes,10,rep,name=FrontendInputs,proto3" json:"FrontendInputs,omitempty" protobuf_key:"bytes,1,opt,name=key" protobuf_val:"bytes,2,opt,name=value"`
	Internal                bool                      `protobuf:"varint,11,opt,name=Internal,proto3" json:"Internal,omitempty"` // Internal builds are not recorded in build history
	SourcePolicy            *pb1.Policy               `protobuf:"bytes,12,opt,name=SourcePolicy,proto3" json:"SourcePolicy,omitempty"`
	Exporters               []*Exporter               `protobuf:"bytes,13,rep,name=Exporters,proto3" json:"Exporters,omitempty"`
	EnableSessionExporter   bool                      `protobuf:"varint,14,opt,name=EnableSessionExporter,proto3" json:"EnableSessionExporter,omitempty"`
	unknownFields           protoimpl.UnknownFields
	sizeCache               protoimpl.SizeCache
}

func (x *SolveRequest) Reset() {
	*x = SolveRequest{}
	mi := &file_github_com_moby_buildkit_api_services_control_control_proto_msgTypes[4]
	ms := protoimpl.X.MessageStateOf(protoimpl.Pointer(x))
	ms.StoreMessageInfo(mi)
}

func (x *SolveRequest) String() string {
	return protoimpl.X.MessageStringOf(x)
}

func (*SolveRequest) ProtoMessage() {}

func (x *SolveRequest) ProtoReflect() protoreflect.Message {
	mi := &file_github_com_moby_buildkit_api_services_control_control_proto_msgTypes[4]
	if x != nil {
		ms := protoimpl.X.MessageStateOf(protoimpl.Pointer(x))
		if ms.LoadMessageInfo() == nil {
			ms.StoreMessageInfo(mi)
		}
		return ms
	}
	return mi.MessageOf(x)
}

// Deprecated: Use SolveRequest.ProtoReflect.Descriptor instead.
func (*SolveRequest) Descriptor() ([]byte, []int) {
	return file_github_com_moby_buildkit_api_services_control_control_proto_rawDescGZIP(), []int{4}
}

func (x *SolveRequest) GetRef() string {
	if x != nil {
		return x.Ref
	}
	return ""
}

func (x *SolveRequest) GetDefinition() *pb.Definition {
	if x != nil {
		return x.Definition
	}
	return nil
}

func (x *SolveRequest) GetExporterDeprecated() string {
	if x != nil {
		return x.ExporterDeprecated
	}
	return ""
}

func (x *SolveRequest) GetExporterAttrsDeprecated() map[string]string {
	if x != nil {
		return x.ExporterAttrsDeprecated
	}
	return nil
}

func (x *SolveRequest) GetSession() string {
	if x != nil {
		return x.Session
	}
	return ""
}

func (x *SolveRequest) GetFrontend() string {
	if x != nil {
		return x.Frontend
	}
	return ""
}

func (x *SolveRequest) GetFrontendAttrs() map[string]string {
	if x != nil {
		return x.FrontendAttrs
	}
	return nil
}

func (x *SolveRequest) GetCache() *CacheOptions {
	if x != nil {
		return x.Cache
	}
	return nil
}

func (x *SolveRequest) GetEntitlements() []string {
	if x != nil {
		return x.Entitlements
	}
	return nil
}

func (x *SolveRequest) GetFrontendInputs() map[string]*pb.Definition {
	if x != nil {
		return x.FrontendInputs
	}
	return nil
}

func (x *SolveRequest) GetInternal() bool {
	if x != nil {
		return x.Internal
	}
	return false
}

func (x *SolveRequest) GetSourcePolicy() *pb1.Policy {
	if x != nil {
		return x.SourcePolicy
	}
	return nil
}

func (x *SolveRequest) GetExporters() []*Exporter {
	if x != nil {
		return x.Exporters
	}
	return nil
}

func (x *SolveRequest) GetEnableSessionExporter() bool {
	if x != nil {
		return x.EnableSessionExporter
	}
	return false
}

type CacheOptions struct {
	state protoimpl.MessageState `protogen:"open.v1"`
	// ExportRefDeprecated is deprecated in favor or the new Exports since BuildKit v0.4.0.
	// When ExportRefDeprecated is set, the solver appends
	// {.Type = "registry", .Attrs = ExportAttrs.add("ref", ExportRef)}
	// to Exports for compatibility. (planned to be removed)
	ExportRefDeprecated string `protobuf:"bytes,1,opt,name=ExportRefDeprecated,proto3" json:"ExportRefDeprecated,omitempty"`
	// ImportRefsDeprecated is deprecated in favor or the new Imports since BuildKit v0.4.0.
	// When ImportRefsDeprecated is set, the solver appends
	// {.Type = "registry", .Attrs = {"ref": importRef}}
	// for each of the ImportRefs entry to Imports for compatibility. (planned to be removed)
	ImportRefsDeprecated []string `protobuf:"bytes,2,rep,name=ImportRefsDeprecated,proto3" json:"ImportRefsDeprecated,omitempty"`
	// ExportAttrsDeprecated is deprecated since BuildKit v0.4.0.
	// See the description of ExportRefDeprecated.
	ExportAttrsDeprecated map[string]string `protobuf:"bytes,3,rep,name=ExportAttrsDeprecated,proto3" json:"ExportAttrsDeprecated,omitempty" protobuf_key:"bytes,1,opt,name=key" protobuf_val:"bytes,2,opt,name=value"`
	// Exports was introduced in BuildKit v0.4.0.
	Exports []*CacheOptionsEntry `protobuf:"bytes,4,rep,name=Exports,proto3" json:"Exports,omitempty"`
	// Imports was introduced in BuildKit v0.4.0.
	Imports       []*CacheOptionsEntry `protobuf:"bytes,5,rep,name=Imports,proto3" json:"Imports,omitempty"`
	unknownFields protoimpl.UnknownFields
	sizeCache     protoimpl.SizeCache
}

func (x *CacheOptions) Reset() {
	*x = CacheOptions{}
	mi := &file_github_com_moby_buildkit_api_services_control_control_proto_msgTypes[5]
	ms := protoimpl.X.MessageStateOf(protoimpl.Pointer(x))
	ms.StoreMessageInfo(mi)
}

func (x *CacheOptions) String() string {
	return protoimpl.X.MessageStringOf(x)
}

func (*CacheOptions) ProtoMessage() {}

func (x *CacheOptions) ProtoReflect() protoreflect.Message {
	mi := &file_github_com_moby_buildkit_api_services_control_control_proto_msgTypes[5]
	if x != nil {
		ms := protoimpl.X.MessageStateOf(protoimpl.Pointer(x))
		if ms.LoadMessageInfo() == nil {
			ms.StoreMessageInfo(mi)
		}
		return ms
	}
	return mi.MessageOf(x)
}

// Deprecated: Use CacheOptions.ProtoReflect.Descriptor instead.
func (*CacheOptions) Descriptor() ([]byte, []int) {
	return file_github_com_moby_buildkit_api_services_control_control_proto_rawDescGZIP(), []int{5}
}

func (x *CacheOptions) GetExportRefDeprecated() string {
	if x != nil {
		return x.ExportRefDeprecated
	}
	return ""
}

func (x *CacheOptions) GetImportRefsDeprecated() []string {
	if x != nil {
		return x.ImportRefsDeprecated
	}
	return nil
}

func (x *CacheOptions) GetExportAttrsDeprecated() map[string]string {
	if x != nil {
		return x.ExportAttrsDeprecated
	}
	return nil
}

func (x *CacheOptions) GetExports() []*CacheOptionsEntry {
	if x != nil {
		return x.Exports
	}
	return nil
}

func (x *CacheOptions) GetImports() []*CacheOptionsEntry {
	if x != nil {
		return x.Imports
	}
	return nil
}

type CacheOptionsEntry struct {
	state protoimpl.MessageState `protogen:"open.v1"`
	// Type is like "registry" or "local"
	Type string `protobuf:"bytes,1,opt,name=Type,proto3" json:"Type,omitempty"`
	// Attrs are like mode=(min,max), ref=example.com:5000/foo/bar .
	// See cache importer/exporter implementations' documentation.
	Attrs         map[string]string `protobuf:"bytes,2,rep,name=Attrs,proto3" json:"Attrs,omitempty" protobuf_key:"bytes,1,opt,name=key" protobuf_val:"bytes,2,opt,name=value"`
	unknownFields protoimpl.UnknownFields
	sizeCache     protoimpl.SizeCache
}

func (x *CacheOptionsEntry) Reset() {
	*x = CacheOptionsEntry{}
	mi := &file_github_com_moby_buildkit_api_services_control_control_proto_msgTypes[6]
	ms := protoimpl.X.MessageStateOf(protoimpl.Pointer(x))
	ms.StoreMessageInfo(mi)
}

func (x *CacheOptionsEntry) String() string {
	return protoimpl.X.MessageStringOf(x)
}

func (*CacheOptionsEntry) ProtoMessage() {}

func (x *CacheOptionsEntry) ProtoReflect() protoreflect.Message {
	mi := &file_github_com_moby_buildkit_api_services_control_control_proto_msgTypes[6]
	if x != nil {
		ms := protoimpl.X.MessageStateOf(protoimpl.Pointer(x))
		if ms.LoadMessageInfo() == nil {
			ms.StoreMessageInfo(mi)
		}
		return ms
	}
	return mi.MessageOf(x)
}

// Deprecated: Use CacheOptionsEntry.ProtoReflect.Descriptor instead.
func (*CacheOptionsEntry) Descriptor() ([]byte, []int) {
	return file_github_com_moby_buildkit_api_services_control_control_proto_rawDescGZIP(), []int{6}
}

func (x *CacheOptionsEntry) GetType() string {
	if x != nil {
		return x.Type
	}
	return ""
}

func (x *CacheOptionsEntry) GetAttrs() map[string]string {
	if x != nil {
		return x.Attrs
	}
	return nil
}

type SolveResponse struct {
	state            protoimpl.MessageState `protogen:"open.v1"`
	ExporterResponse map[string]string      `protobuf:"bytes,1,rep,name=ExporterResponse,proto3" json:"ExporterResponse,omitempty" protobuf_key:"bytes,1,opt,name=key" protobuf_val:"bytes,2,opt,name=value"`
	unknownFields    protoimpl.UnknownFields
	sizeCache        protoimpl.SizeCache
}

func (x *SolveResponse) Reset() {
	*x = SolveResponse{}
	mi := &file_github_com_moby_buildkit_api_services_control_control_proto_msgTypes[7]
	ms := protoimpl.X.MessageStateOf(protoimpl.Pointer(x))
	ms.StoreMessageInfo(mi)
}

func (x *SolveResponse) String() string {
	return protoimpl.X.MessageStringOf(x)
}

func (*SolveResponse) ProtoMessage() {}

func (x *SolveResponse) ProtoReflect() protoreflect.Message {
	mi := &file_github_com_moby_buildkit_api_services_control_control_proto_msgTypes[7]
	if x != nil {
		ms := protoimpl.X.MessageStateOf(protoimpl.Pointer(x))
		if ms.LoadMessageInfo() == nil {
			ms.StoreMessageInfo(mi)
		}
		return ms
	}
	return mi.MessageOf(x)
}

// Deprecated: Use SolveResponse.ProtoReflect.Descriptor instead.
func (*SolveResponse) Descriptor() ([]byte, []int) {
	return file_github_com_moby_buildkit_api_services_control_control_proto_rawDescGZIP(), []int{7}
}

func (x *SolveResponse) GetExporterResponse() map[string]string {
	if x != nil {
		return x.ExporterResponse
	}
	return nil
}

type StatusRequest struct {
	state         protoimpl.MessageState `protogen:"open.v1"`
	Ref           string                 `protobuf:"bytes,1,opt,name=Ref,proto3" json:"Ref,omitempty"`
	unknownFields protoimpl.UnknownFields
	sizeCache     protoimpl.SizeCache
}

func (x *StatusRequest) Reset() {
	*x = StatusRequest{}
	mi := &file_github_com_moby_buildkit_api_services_control_control_proto_msgTypes[8]
	ms := protoimpl.X.MessageStateOf(protoimpl.Pointer(x))
	ms.StoreMessageInfo(mi)
}

func (x *StatusRequest) String() string {
	return protoimpl.X.MessageStringOf(x)
}

func (*StatusRequest) ProtoMessage() {}

func (x *StatusRequest) ProtoReflect() protoreflect.Message {
	mi := &file_github_com_moby_buildkit_api_services_control_control_proto_msgTypes[8]
	if x != nil {
		ms := protoimpl.X.MessageStateOf(protoimpl.Pointer(x))
		if ms.LoadMessageInfo() == nil {
			ms.StoreMessageInfo(mi)
		}
		return ms
	}
	return mi.MessageOf(x)
}

// Deprecated: Use StatusRequest.ProtoReflect.Descriptor instead.
func (*StatusRequest) Descriptor() ([]byte, []int) {
	return file_github_com_moby_buildkit_api_services_control_control_proto_rawDescGZIP(), []int{8}
}

func (x *StatusRequest) GetRef() string {
	if x != nil {
		return x.Ref
	}
	return ""
}

type StatusResponse struct {
	state         protoimpl.MessageState `protogen:"open.v1"`
	Vertexes      []*Vertex              `protobuf:"bytes,1,rep,name=vertexes,proto3" json:"vertexes,omitempty"`
	Statuses      []*VertexStatus        `protobuf:"bytes,2,rep,name=statuses,proto3" json:"statuses,omitempty"`
	Logs          []*VertexLog           `protobuf:"bytes,3,rep,name=logs,proto3" json:"logs,omitempty"`
	Warnings      []*VertexWarning       `protobuf:"bytes,4,rep,name=warnings,proto3" json:"warnings,omitempty"`
	unknownFields protoimpl.UnknownFields
	sizeCache     protoimpl.SizeCache
}

func (x *StatusResponse) Reset() {
	*x = StatusResponse{}
	mi := &file_github_com_moby_buildkit_api_services_control_control_proto_msgTypes[9]
	ms := protoimpl.X.MessageStateOf(protoimpl.Pointer(x))
	ms.StoreMessageInfo(mi)
}

func (x *StatusResponse) String() string {
	return protoimpl.X.MessageStringOf(x)
}

func (*StatusResponse) ProtoMessage() {}

func (x *StatusResponse) ProtoReflect() protoreflect.Message {
	mi := &file_github_com_moby_buildkit_api_services_control_control_proto_msgTypes[9]
	if x != nil {
		ms := protoimpl.X.MessageStateOf(protoimpl.Pointer(x))
		if ms.LoadMessageInfo() == nil {
			ms.StoreMessageInfo(mi)
		}
		return ms
	}
	return mi.MessageOf(x)
}

// Deprecated: Use StatusResponse.ProtoReflect.Descriptor instead.
func (*StatusResponse) Descriptor() ([]byte, []int) {
	return file_github_com_moby_buildkit_api_services_control_control_proto_rawDescGZIP(), []int{9}
}

func (x *StatusResponse) GetVertexes() []*Vertex {
	if x != nil {
		return x.Vertexes
	}
	return nil
}

func (x *StatusResponse) GetStatuses() []*VertexStatus {
	if x != nil {
		return x.Statuses
	}
	return nil
}

func (x *StatusResponse) GetLogs() []*VertexLog {
	if x != nil {
		return x.Logs
	}
	return nil
}

func (x *StatusResponse) GetWarnings() []*VertexWarning {
	if x != nil {
		return x.Warnings
	}
	return nil
}

type Vertex struct {
	state         protoimpl.MessageState `protogen:"open.v1"`
	Digest        string                 `protobuf:"bytes,1,opt,name=digest,proto3" json:"digest,omitempty"`
	Inputs        []string               `protobuf:"bytes,2,rep,name=inputs,proto3" json:"inputs,omitempty"`
	Name          string                 `protobuf:"bytes,3,opt,name=name,proto3" json:"name,omitempty"`
	Cached        bool                   `protobuf:"varint,4,opt,name=cached,proto3" json:"cached,omitempty"`
	Started       *timestamp.Timestamp   `protobuf:"bytes,5,opt,name=started,proto3" json:"started,omitempty"`
	Completed     *timestamp.Timestamp   `protobuf:"bytes,6,opt,name=completed,proto3" json:"completed,omitempty"`
	Error         string                 `protobuf:"bytes,7,opt,name=error,proto3" json:"error,omitempty"` // typed errors?
	ProgressGroup *pb.ProgressGroup      `protobuf:"bytes,8,opt,name=progressGroup,proto3" json:"progressGroup,omitempty"`
	unknownFields protoimpl.UnknownFields
	sizeCache     protoimpl.SizeCache
}

func (x *Vertex) Reset() {
	*x = Vertex{}
	mi := &file_github_com_moby_buildkit_api_services_control_control_proto_msgTypes[10]
	ms := protoimpl.X.MessageStateOf(protoimpl.Pointer(x))
	ms.StoreMessageInfo(mi)
}

func (x *Vertex) String() string {
	return protoimpl.X.MessageStringOf(x)
}

func (*Vertex) ProtoMessage() {}

func (x *Vertex) ProtoReflect() protoreflect.Message {
	mi := &file_github_com_moby_buildkit_api_services_control_control_proto_msgTypes[10]
	if x != nil {
		ms := protoimpl.X.MessageStateOf(protoimpl.Pointer(x))
		if ms.LoadMessageInfo() == nil {
			ms.StoreMessageInfo(mi)
		}
		return ms
	}
	return mi.MessageOf(x)
}

// Deprecated: Use Vertex.ProtoReflect.Descriptor instead.
func (*Vertex) Descriptor() ([]byte, []int) {
	return file_github_com_moby_buildkit_api_services_control_control_proto_rawDescGZIP(), []int{10}
}

func (x *Vertex) GetDigest() string {
	if x != nil {
		return x.Digest
	}
	return ""
}

func (x *Vertex) GetInputs() []string {
	if x != nil {
		return x.Inputs
	}
	return nil
}

func (x *Vertex) GetName() string {
	if x != nil {
		return x.Name
	}
	return ""
}

func (x *Vertex) GetCached() bool {
	if x != nil {
		return x.Cached
	}
	return false
}

func (x *Vertex) GetStarted() *timestamp.Timestamp {
	if x != nil {
		return x.Started
	}
	return nil
}

func (x *Vertex) GetCompleted() *timestamp.Timestamp {
	if x != nil {
		return x.Completed
	}
	return nil
}

func (x *Vertex) GetError() string {
	if x != nil {
		return x.Error
	}
	return ""
}

func (x *Vertex) GetProgressGroup() *pb.ProgressGroup {
	if x != nil {
		return x.ProgressGroup
	}
	return nil
}

type VertexStatus struct {
	state         protoimpl.MessageState `protogen:"open.v1"`
	ID            string                 `protobuf:"bytes,1,opt,name=ID,proto3" json:"ID,omitempty"`
	Vertex        string                 `protobuf:"bytes,2,opt,name=vertex,proto3" json:"vertex,omitempty"`
	Name          string                 `protobuf:"bytes,3,opt,name=name,proto3" json:"name,omitempty"`
	Current       int64                  `protobuf:"varint,4,opt,name=current,proto3" json:"current,omitempty"`
	Total         int64                  `protobuf:"varint,5,opt,name=total,proto3" json:"total,omitempty"`
	Timestamp     *timestamp.Timestamp   `protobuf:"bytes,6,opt,name=timestamp,proto3" json:"timestamp,omitempty"`
	Started       *timestamp.Timestamp   `protobuf:"bytes,7,opt,name=started,proto3" json:"started,omitempty"`
	Completed     *timestamp.Timestamp   `protobuf:"bytes,8,opt,name=completed,proto3" json:"completed,omitempty"`
	unknownFields protoimpl.UnknownFields
	sizeCache     protoimpl.SizeCache
}

func (x *VertexStatus) Reset() {
	*x = VertexStatus{}
	mi := &file_github_com_moby_buildkit_api_services_control_control_proto_msgTypes[11]
	ms := protoimpl.X.MessageStateOf(protoimpl.Pointer(x))
	ms.StoreMessageInfo(mi)
}

func (x *VertexStatus) String() string {
	return protoimpl.X.MessageStringOf(x)
}

func (*VertexStatus) ProtoMessage() {}

func (x *VertexStatus) ProtoReflect() protoreflect.Message {
	mi := &file_github_com_moby_buildkit_api_services_control_control_proto_msgTypes[11]
	if x != nil {
		ms := protoimpl.X.MessageStateOf(protoimpl.Pointer(x))
		if ms.LoadMessageInfo() == nil {
			ms.StoreMessageInfo(mi)
		}
		return ms
	}
	return mi.MessageOf(x)
}

// Deprecated: Use VertexStatus.ProtoReflect.Descriptor instead.
func (*VertexStatus) Descriptor() ([]byte, []int) {
	return file_github_com_moby_buildkit_api_services_control_control_proto_rawDescGZIP(), []int{11}
}

func (x *VertexStatus) GetID() string {
	if x != nil {
		return x.ID
	}
	return ""
}

func (x *VertexStatus) GetVertex() string {
	if x != nil {
		return x.Vertex
	}
	return ""
}

func (x *VertexStatus) GetName() string {
	if x != nil {
		return x.Name
	}
	return ""
}

func (x *VertexStatus) GetCurrent() int64 {
	if x != nil {
		return x.Current
	}
	return 0
}

func (x *VertexStatus) GetTotal() int64 {
	if x != nil {
		return x.Total
	}
	return 0
}

func (x *VertexStatus) GetTimestamp() *timestamp.Timestamp {
	if x != nil {
		return x.Timestamp
	}
	return nil
}

func (x *VertexStatus) GetStarted() *timestamp.Timestamp {
	if x != nil {
		return x.Started
	}
	return nil
}

func (x *VertexStatus) GetCompleted() *timestamp.Timestamp {
	if x != nil {
		return x.Completed
	}
	return nil
}

type VertexLog struct {
	state         protoimpl.MessageState `protogen:"open.v1"`
	Vertex        string                 `protobuf:"bytes,1,opt,name=vertex,proto3" json:"vertex,omitempty"`
	Timestamp     *timestamp.Timestamp   `protobuf:"bytes,2,opt,name=timestamp,proto3" json:"timestamp,omitempty"`
	Stream        int64                  `protobuf:"varint,3,opt,name=stream,proto3" json:"stream,omitempty"`
	Msg           []byte                 `protobuf:"bytes,4,opt,name=msg,proto3" json:"msg,omitempty"`
	unknownFields protoimpl.UnknownFields
	sizeCache     protoimpl.SizeCache
}

func (x *VertexLog) Reset() {
	*x = VertexLog{}
	mi := &file_github_com_moby_buildkit_api_services_control_control_proto_msgTypes[12]
	ms := protoimpl.X.MessageStateOf(protoimpl.Pointer(x))
	ms.StoreMessageInfo(mi)
}

func (x *VertexLog) String() string {
	return protoimpl.X.MessageStringOf(x)
}

func (*VertexLog) ProtoMessage() {}

func (x *VertexLog) ProtoReflect() protoreflect.Message {
	mi := &file_github_com_moby_buildkit_api_services_control_control_proto_msgTypes[12]
	if x != nil {
		ms := protoimpl.X.MessageStateOf(protoimpl.Pointer(x))
		if ms.LoadMessageInfo() == nil {
			ms.StoreMessageInfo(mi)
		}
		return ms
	}
	return mi.MessageOf(x)
}

// Deprecated: Use VertexLog.ProtoReflect.Descriptor instead.
func (*VertexLog) Descriptor() ([]byte, []int) {
	return file_github_com_moby_buildkit_api_services_control_control_proto_rawDescGZIP(), []int{12}
}

func (x *VertexLog) GetVertex() string {
	if x != nil {
		return x.Vertex
	}
	return ""
}

func (x *VertexLog) GetTimestamp() *timestamp.Timestamp {
	if x != nil {
		return x.Timestamp
	}
	return nil
}

func (x *VertexLog) GetStream() int64 {
	if x != nil {
		return x.Stream
	}
	return 0
}

func (x *VertexLog) GetMsg() []byte {
	if x != nil {
		return x.Msg
	}
	return nil
}

type VertexWarning struct {
	state         protoimpl.MessageState `protogen:"open.v1"`
	Vertex        string                 `protobuf:"bytes,1,opt,name=vertex,proto3" json:"vertex,omitempty"`
	Level         int64                  `protobuf:"varint,2,opt,name=level,proto3" json:"level,omitempty"`
	Short         []byte                 `protobuf:"bytes,3,opt,name=short,proto3" json:"short,omitempty"`
	Detail        [][]byte               `protobuf:"bytes,4,rep,name=detail,proto3" json:"detail,omitempty"`
	Url           string                 `protobuf:"bytes,5,opt,name=url,proto3" json:"url,omitempty"`
	Info          *pb.SourceInfo         `protobuf:"bytes,6,opt,name=info,proto3" json:"info,omitempty"`
	Ranges        []*pb.Range            `protobuf:"bytes,7,rep,name=ranges,proto3" json:"ranges,omitempty"`
	unknownFields protoimpl.UnknownFields
	sizeCache     protoimpl.SizeCache
}

func (x *VertexWarning) Reset() {
	*x = VertexWarning{}
	mi := &file_github_com_moby_buildkit_api_services_control_control_proto_msgTypes[13]
	ms := protoimpl.X.MessageStateOf(protoimpl.Pointer(x))
	ms.StoreMessageInfo(mi)
}

func (x *VertexWarning) String() string {
	return protoimpl.X.MessageStringOf(x)
}

func (*VertexWarning) ProtoMessage() {}

func (x *VertexWarning) ProtoReflect() protoreflect.Message {
	mi := &file_github_com_moby_buildkit_api_services_control_control_proto_msgTypes[13]
	if x != nil {
		ms := protoimpl.X.MessageStateOf(protoimpl.Pointer(x))
		if ms.LoadMessageInfo() == nil {
			ms.StoreMessageInfo(mi)
		}
		return ms
	}
	return mi.MessageOf(x)
}

// Deprecated: Use VertexWarning.ProtoReflect.Descriptor instead.
func (*VertexWarning) Descriptor() ([]byte, []int) {
	return file_github_com_moby_buildkit_api_services_control_control_proto_rawDescGZIP(), []int{13}
}

func (x *VertexWarning) GetVertex() string {
	if x != nil {
		return x.Vertex
	}
	return ""
}

func (x *VertexWarning) GetLevel() int64 {
	if x != nil {
		return x.Level
	}
	return 0
}

func (x *VertexWarning) GetShort() []byte {
	if x != nil {
		return x.Short
	}
	return nil
}

func (x *VertexWarning) GetDetail() [][]byte {
	if x != nil {
		return x.Detail
	}
	return nil
}

func (x *VertexWarning) GetUrl() string {
	if x != nil {
		return x.Url
	}
	return ""
}

func (x *VertexWarning) GetInfo() *pb.SourceInfo {
	if x != nil {
		return x.Info
	}
	return nil
}

func (x *VertexWarning) GetRanges() []*pb.Range {
	if x != nil {
		return x.Ranges
	}
	return nil
}

type BytesMessage struct {
	state         protoimpl.MessageState `protogen:"open.v1"`
	Data          []byte                 `protobuf:"bytes,1,opt,name=data,proto3" json:"data,omitempty"`
	unknownFields protoimpl.UnknownFields
	sizeCache     protoimpl.SizeCache
}

func (x *BytesMessage) Reset() {
	*x = BytesMessage{}
	mi := &file_github_com_moby_buildkit_api_services_control_control_proto_msgTypes[14]
	ms := protoimpl.X.MessageStateOf(protoimpl.Pointer(x))
	ms.StoreMessageInfo(mi)
}

func (x *BytesMessage) String() string {
	return protoimpl.X.MessageStringOf(x)
}

func (*BytesMessage) ProtoMessage() {}

func (x *BytesMessage) ProtoReflect() protoreflect.Message {
	mi := &file_github_com_moby_buildkit_api_services_control_control_proto_msgTypes[14]
	if x != nil {
		ms := protoimpl.X.MessageStateOf(protoimpl.Pointer(x))
		if ms.LoadMessageInfo() == nil {
			ms.StoreMessageInfo(mi)
		}
		return ms
	}
	return mi.MessageOf(x)
}

// Deprecated: Use BytesMessage.ProtoReflect.Descriptor instead.
func (*BytesMessage) Descriptor() ([]byte, []int) {
	return file_github_com_moby_buildkit_api_services_control_control_proto_rawDescGZIP(), []int{14}
}

func (x *BytesMessage) GetData() []byte {
	if x != nil {
		return x.Data
	}
	return nil
}

type ListWorkersRequest struct {
	state         protoimpl.MessageState `protogen:"open.v1"`
	Filter        []string               `protobuf:"bytes,1,rep,name=filter,proto3" json:"filter,omitempty"` // containerd style
	unknownFields protoimpl.UnknownFields
	sizeCache     protoimpl.SizeCache
}

func (x *ListWorkersRequest) Reset() {
	*x = ListWorkersRequest{}
	mi := &file_github_com_moby_buildkit_api_services_control_control_proto_msgTypes[15]
	ms := protoimpl.X.MessageStateOf(protoimpl.Pointer(x))
	ms.StoreMessageInfo(mi)
}

func (x *ListWorkersRequest) String() string {
	return protoimpl.X.MessageStringOf(x)
}

func (*ListWorkersRequest) ProtoMessage() {}

func (x *ListWorkersRequest) ProtoReflect() protoreflect.Message {
	mi := &file_github_com_moby_buildkit_api_services_control_control_proto_msgTypes[15]
	if x != nil {
		ms := protoimpl.X.MessageStateOf(protoimpl.Pointer(x))
		if ms.LoadMessageInfo() == nil {
			ms.StoreMessageInfo(mi)
		}
		return ms
	}
	return mi.MessageOf(x)
}

// Deprecated: Use ListWorkersRequest.ProtoReflect.Descriptor instead.
func (*ListWorkersRequest) Descriptor() ([]byte, []int) {
	return file_github_com_moby_buildkit_api_services_control_control_proto_rawDescGZIP(), []int{15}
}

func (x *ListWorkersRequest) GetFilter() []string {
	if x != nil {
		return x.Filter
	}
	return nil
}

type ListWorkersResponse struct {
	state         protoimpl.MessageState `protogen:"open.v1"`
	Record        []*types.WorkerRecord  `protobuf:"bytes,1,rep,name=record,proto3" json:"record,omitempty"`
	unknownFields protoimpl.UnknownFields
	sizeCache     protoimpl.SizeCache
}

func (x *ListWorkersResponse) Reset() {
	*x = ListWorkersResponse{}
	mi := &file_github_com_moby_buildkit_api_services_control_control_proto_msgTypes[16]
	ms := protoimpl.X.MessageStateOf(protoimpl.Pointer(x))
	ms.StoreMessageInfo(mi)
}

func (x *ListWorkersResponse) String() string {
	return protoimpl.X.MessageStringOf(x)
}

func (*ListWorkersResponse) ProtoMessage() {}

func (x *ListWorkersResponse) ProtoReflect() protoreflect.Message {
	mi := &file_github_com_moby_buildkit_api_services_control_control_proto_msgTypes[16]
	if x != nil {
		ms := protoimpl.X.MessageStateOf(protoimpl.Pointer(x))
		if ms.LoadMessageInfo() == nil {
			ms.StoreMessageInfo(mi)
		}
		return ms
	}
	return mi.MessageOf(x)
}

// Deprecated: Use ListWorkersResponse.ProtoReflect.Descriptor instead.
func (*ListWorkersResponse) Descriptor() ([]byte, []int) {
	return file_github_com_moby_buildkit_api_services_control_control_proto_rawDescGZIP(), []int{16}
}

func (x *ListWorkersResponse) GetRecord() []*types.WorkerRecord {
	if x != nil {
		return x.Record
	}
	return nil
}

type InfoRequest struct {
	state         protoimpl.MessageState `protogen:"open.v1"`
	unknownFields protoimpl.UnknownFields
	sizeCache     protoimpl.SizeCache
}

func (x *InfoRequest) Reset() {
	*x = InfoRequest{}
	mi := &file_github_com_moby_buildkit_api_services_control_control_proto_msgTypes[17]
	ms := protoimpl.X.MessageStateOf(protoimpl.Pointer(x))
	ms.StoreMessageInfo(mi)
}

func (x *InfoRequest) String() string {
	return protoimpl.X.MessageStringOf(x)
}

func (*InfoRequest) ProtoMessage() {}

func (x *InfoRequest) ProtoReflect() protoreflect.Message {
	mi := &file_github_com_moby_buildkit_api_services_control_control_proto_msgTypes[17]
	if x != nil {
		ms := protoimpl.X.MessageStateOf(protoimpl.Pointer(x))
		if ms.LoadMessageInfo() == nil {
			ms.StoreMessageInfo(mi)
		}
		return ms
	}
	return mi.MessageOf(x)
}

// Deprecated: Use InfoRequest.ProtoReflect.Descriptor instead.
func (*InfoRequest) Descriptor() ([]byte, []int) {
	return file_github_com_moby_buildkit_api_services_control_control_proto_rawDescGZIP(), []int{17}
}

type InfoResponse struct {
	state           protoimpl.MessageState `protogen:"open.v1"`
	BuildkitVersion *types.BuildkitVersion `protobuf:"bytes,1,opt,name=buildkitVersion,proto3" json:"buildkitVersion,omitempty"`
	unknownFields   protoimpl.UnknownFields
	sizeCache       protoimpl.SizeCache
}

func (x *InfoResponse) Reset() {
	*x = InfoResponse{}
	mi := &file_github_com_moby_buildkit_api_services_control_control_proto_msgTypes[18]
	ms := protoimpl.X.MessageStateOf(protoimpl.Pointer(x))
	ms.StoreMessageInfo(mi)
}

func (x *InfoResponse) String() string {
	return protoimpl.X.MessageStringOf(x)
}

func (*InfoResponse) ProtoMessage() {}

func (x *InfoResponse) ProtoReflect() protoreflect.Message {
	mi := &file_github_com_moby_buildkit_api_services_control_control_proto_msgTypes[18]
	if x != nil {
		ms := protoimpl.X.MessageStateOf(protoimpl.Pointer(x))
		if ms.LoadMessageInfo() == nil {
			ms.StoreMessageInfo(mi)
		}
		return ms
	}
	return mi.MessageOf(x)
}

// Deprecated: Use InfoResponse.ProtoReflect.Descriptor instead.
func (*InfoResponse) Descriptor() ([]byte, []int) {
	return file_github_com_moby_buildkit_api_services_control_control_proto_rawDescGZIP(), []int{18}
}

func (x *InfoResponse) GetBuildkitVersion() *types.BuildkitVersion {
	if x != nil {
		return x.BuildkitVersion
	}
	return nil
}

type BuildHistoryRequest struct {
	state         protoimpl.MessageState `protogen:"open.v1"`
	ActiveOnly    bool                   `protobuf:"varint,1,opt,name=ActiveOnly,proto3" json:"ActiveOnly,omitempty"`
	Ref           string                 `protobuf:"bytes,2,opt,name=Ref,proto3" json:"Ref,omitempty"`
	EarlyExit     bool                   `protobuf:"varint,3,opt,name=EarlyExit,proto3" json:"EarlyExit,omitempty"`
	Filter        []string               `protobuf:"bytes,4,rep,name=Filter,proto3" json:"Filter,omitempty"`
	Limit         int32                  `protobuf:"varint,5,opt,name=Limit,proto3" json:"Limit,omitempty"`
	unknownFields protoimpl.UnknownFields
	sizeCache     protoimpl.SizeCache
}

func (x *BuildHistoryRequest) Reset() {
	*x = BuildHistoryRequest{}
	mi := &file_github_com_moby_buildkit_api_services_control_control_proto_msgTypes[19]
	ms := protoimpl.X.MessageStateOf(protoimpl.Pointer(x))
	ms.StoreMessageInfo(mi)
}

func (x *BuildHistoryRequest) String() string {
	return protoimpl.X.MessageStringOf(x)
}

func (*BuildHistoryRequest) ProtoMessage() {}

func (x *BuildHistoryRequest) ProtoReflect() protoreflect.Message {
	mi := &file_github_com_moby_buildkit_api_services_control_control_proto_msgTypes[19]
	if x != nil {
		ms := protoimpl.X.MessageStateOf(protoimpl.Pointer(x))
		if ms.LoadMessageInfo() == nil {
			ms.StoreMessageInfo(mi)
		}
		return ms
	}
	return mi.MessageOf(x)
}

// Deprecated: Use BuildHistoryRequest.ProtoReflect.Descriptor instead.
func (*BuildHistoryRequest) Descriptor() ([]byte, []int) {
	return file_github_com_moby_buildkit_api_services_control_control_proto_rawDescGZIP(), []int{19}
}

func (x *BuildHistoryRequest) GetActiveOnly() bool {
	if x != nil {
		return x.ActiveOnly
	}
	return false
}

func (x *BuildHistoryRequest) GetRef() string {
	if x != nil {
		return x.Ref
	}
	return ""
}

func (x *BuildHistoryRequest) GetEarlyExit() bool {
	if x != nil {
		return x.EarlyExit
	}
	return false
}

func (x *BuildHistoryRequest) GetFilter() []string {
	if x != nil {
		return x.Filter
	}
	return nil
}

func (x *BuildHistoryRequest) GetLimit() int32 {
	if x != nil {
		return x.Limit
	}
	return 0
}

type BuildHistoryEvent struct {
	state         protoimpl.MessageState `protogen:"open.v1"`
	Type          BuildHistoryEventType  `protobuf:"varint,1,opt,name=type,proto3,enum=moby.buildkit.v1.BuildHistoryEventType" json:"type,omitempty"`
	Record        *BuildHistoryRecord    `protobuf:"bytes,2,opt,name=record,proto3" json:"record,omitempty"`
	unknownFields protoimpl.UnknownFields
	sizeCache     protoimpl.SizeCache
}

func (x *BuildHistoryEvent) Reset() {
	*x = BuildHistoryEvent{}
	mi := &file_github_com_moby_buildkit_api_services_control_control_proto_msgTypes[20]
	ms := protoimpl.X.MessageStateOf(protoimpl.Pointer(x))
	ms.StoreMessageInfo(mi)
}

func (x *BuildHistoryEvent) String() string {
	return protoimpl.X.MessageStringOf(x)
}

func (*BuildHistoryEvent) ProtoMessage() {}

func (x *BuildHistoryEvent) ProtoReflect() protoreflect.Message {
	mi := &file_github_com_moby_buildkit_api_services_control_control_proto_msgTypes[20]
	if x != nil {
		ms := protoimpl.X.MessageStateOf(protoimpl.Pointer(x))
		if ms.LoadMessageInfo() == nil {
			ms.StoreMessageInfo(mi)
		}
		return ms
	}
	return mi.MessageOf(x)
}

// Deprecated: Use BuildHistoryEvent.ProtoReflect.Descriptor instead.
func (*BuildHistoryEvent) Descriptor() ([]byte, []int) {
	return file_github_com_moby_buildkit_api_services_control_control_proto_rawDescGZIP(), []int{20}
}

func (x *BuildHistoryEvent) GetType() BuildHistoryEventType {
	if x != nil {
		return x.Type
	}
	return BuildHistoryEventType_STARTED
}

func (x *BuildHistoryEvent) GetRecord() *BuildHistoryRecord {
	if x != nil {
		return x.Record
	}
	return nil
}

type BuildHistoryRecord struct {
	state             protoimpl.MessageState      `protogen:"open.v1"`
	Ref               string                      `protobuf:"bytes,1,opt,name=Ref,proto3" json:"Ref,omitempty"`
	Frontend          string                      `protobuf:"bytes,2,opt,name=Frontend,proto3" json:"Frontend,omitempty"`
	FrontendAttrs     map[string]string           `protobuf:"bytes,3,rep,name=FrontendAttrs,proto3" json:"FrontendAttrs,omitempty" protobuf_key:"bytes,1,opt,name=key" protobuf_val:"bytes,2,opt,name=value"`
	Exporters         []*Exporter                 `protobuf:"bytes,4,rep,name=Exporters,proto3" json:"Exporters,omitempty"`
	Error             *status.Status              `protobuf:"bytes,5,opt,name=error,proto3" json:"error,omitempty"`
	CreatedAt         *timestamp.Timestamp        `protobuf:"bytes,6,opt,name=CreatedAt,proto3" json:"CreatedAt,omitempty"`
	CompletedAt       *timestamp.Timestamp        `protobuf:"bytes,7,opt,name=CompletedAt,proto3" json:"CompletedAt,omitempty"`
	Logs              *Descriptor                 `protobuf:"bytes,8,opt,name=logs,proto3" json:"logs,omitempty"`
	ExporterResponse  map[string]string           `protobuf:"bytes,9,rep,name=ExporterResponse,proto3" json:"ExporterResponse,omitempty" protobuf_key:"bytes,1,opt,name=key" protobuf_val:"bytes,2,opt,name=value"`
	Result            *BuildResultInfo            `protobuf:"bytes,10,opt,name=Result,proto3" json:"Result,omitempty"`
	Results           map[string]*BuildResultInfo `protobuf:"bytes,11,rep,name=Results,proto3" json:"Results,omitempty" protobuf_key:"bytes,1,opt,name=key" protobuf_val:"bytes,2,opt,name=value"`
	Generation        int32                       `protobuf:"varint,12,opt,name=Generation,proto3" json:"Generation,omitempty"`
	Trace             *Descriptor                 `protobuf:"bytes,13,opt,name=trace,proto3" json:"trace,omitempty"`
	Pinned            bool                        `protobuf:"varint,14,opt,name=pinned,proto3" json:"pinned,omitempty"`
	NumCachedSteps    int32                       `protobuf:"varint,15,opt,name=numCachedSteps,proto3" json:"numCachedSteps,omitempty"`
	NumTotalSteps     int32                       `protobuf:"varint,16,opt,name=numTotalSteps,proto3" json:"numTotalSteps,omitempty"`
	NumCompletedSteps int32                       `protobuf:"varint,17,opt,name=numCompletedSteps,proto3" json:"numCompletedSteps,omitempty"`
	ExternalError     *Descriptor                 `protobuf:"bytes,18,opt,name=externalError,proto3" json:"externalError,omitempty"`
	NumWarnings       int32                       `protobuf:"varint,19,opt,name=numWarnings,proto3" json:"numWarnings,omitempty"`
	unknownFields     protoimpl.UnknownFields
	sizeCache         protoimpl.SizeCache
}

func (x *BuildHistoryRecord) Reset() {
	*x = BuildHistoryRecord{}
	mi := &file_github_com_moby_buildkit_api_services_control_control_proto_msgTypes[21]
	ms := protoimpl.X.MessageStateOf(protoimpl.Pointer(x))
	ms.StoreMessageInfo(mi)
}

func (x *BuildHistoryRecord) String() string {
	return protoimpl.X.MessageStringOf(x)
}

func (*BuildHistoryRecord) ProtoMessage() {}

func (x *BuildHistoryRecord) ProtoReflect() protoreflect.Message {
	mi := &file_github_com_moby_buildkit_api_services_control_control_proto_msgTypes[21]
	if x != nil {
		ms := protoimpl.X.MessageStateOf(protoimpl.Pointer(x))
		if ms.LoadMessageInfo() == nil {
			ms.StoreMessageInfo(mi)
		}
		return ms
	}
	return mi.MessageOf(x)
}

// Deprecated: Use BuildHistoryRecord.ProtoReflect.Descriptor instead.
func (*BuildHistoryRecord) Descriptor() ([]byte, []int) {
	return file_github_com_moby_buildkit_api_services_control_control_proto_rawDescGZIP(), []int{21}
}

func (x *BuildHistoryRecord) GetRef() string {
	if x != nil {
		return x.Ref
	}
	return ""
}

func (x *BuildHistoryRecord) GetFrontend() string {
	if x != nil {
		return x.Frontend
	}
	return ""
}

func (x *BuildHistoryRecord) GetFrontendAttrs() map[string]string {
	if x != nil {
		return x.FrontendAttrs
	}
	return nil
}

func (x *BuildHistoryRecord) GetExporters() []*Exporter {
	if x != nil {
		return x.Exporters
	}
	return nil
}

func (x *BuildHistoryRecord) GetError() *status.Status {
	if x != nil {
		return x.Error
	}
	return nil
}

func (x *BuildHistoryRecord) GetCreatedAt() *timestamp.Timestamp {
	if x != nil {
		return x.CreatedAt
	}
	return nil
}

func (x *BuildHistoryRecord) GetCompletedAt() *timestamp.Timestamp {
	if x != nil {
		return x.CompletedAt
	}
	return nil
}

func (x *BuildHistoryRecord) GetLogs() *Descriptor {
	if x != nil {
		return x.Logs
	}
	return nil
}

func (x *BuildHistoryRecord) GetExporterResponse() map[string]string {
	if x != nil {
		return x.ExporterResponse
	}
	return nil
}

func (x *BuildHistoryRecord) GetResult() *BuildResultInfo {
	if x != nil {
		return x.Result
	}
	return nil
}

func (x *BuildHistoryRecord) GetResults() map[string]*BuildResultInfo {
	if x != nil {
		return x.Results
	}
	return nil
}

func (x *BuildHistoryRecord) GetGeneration() int32 {
	if x != nil {
		return x.Generation
	}
	return 0
}

func (x *BuildHistoryRecord) GetTrace() *Descriptor {
	if x != nil {
		return x.Trace
	}
	return nil
}

func (x *BuildHistoryRecord) GetPinned() bool {
	if x != nil {
		return x.Pinned
	}
	return false
}

func (x *BuildHistoryRecord) GetNumCachedSteps() int32 {
	if x != nil {
		return x.NumCachedSteps
	}
	return 0
}

func (x *BuildHistoryRecord) GetNumTotalSteps() int32 {
	if x != nil {
		return x.NumTotalSteps
	}
	return 0
}

func (x *BuildHistoryRecord) GetNumCompletedSteps() int32 {
	if x != nil {
		return x.NumCompletedSteps
	}
	return 0
}

func (x *BuildHistoryRecord) GetExternalError() *Descriptor {
	if x != nil {
		return x.ExternalError
	}
	return nil
}

func (x *BuildHistoryRecord) GetNumWarnings() int32 {
	if x != nil {
		return x.NumWarnings
	}
	return 0
}

type UpdateBuildHistoryRequest struct {
	state         protoimpl.MessageState `protogen:"open.v1"`
	Ref           string                 `protobuf:"bytes,1,opt,name=Ref,proto3" json:"Ref,omitempty"`
	Pinned        bool                   `protobuf:"varint,2,opt,name=Pinned,proto3" json:"Pinned,omitempty"`
	Delete        bool                   `protobuf:"varint,3,opt,name=Delete,proto3" json:"Delete,omitempty"`
	Finalize      bool                   `protobuf:"varint,4,opt,name=Finalize,proto3" json:"Finalize,omitempty"`
	unknownFields protoimpl.UnknownFields
	sizeCache     protoimpl.SizeCache
}

func (x *UpdateBuildHistoryRequest) Reset() {
	*x = UpdateBuildHistoryRequest{}
	mi := &file_github_com_moby_buildkit_api_services_control_control_proto_msgTypes[22]
	ms := protoimpl.X.MessageStateOf(protoimpl.Pointer(x))
	ms.StoreMessageInfo(mi)
}

func (x *UpdateBuildHistoryRequest) String() string {
	return protoimpl.X.MessageStringOf(x)
}

func (*UpdateBuildHistoryRequest) ProtoMessage() {}

func (x *UpdateBuildHistoryRequest) ProtoReflect() protoreflect.Message {
	mi := &file_github_com_moby_buildkit_api_services_control_control_proto_msgTypes[22]
	if x != nil {
		ms := protoimpl.X.MessageStateOf(protoimpl.Pointer(x))
		if ms.LoadMessageInfo() == nil {
			ms.StoreMessageInfo(mi)
		}
		return ms
	}
	return mi.MessageOf(x)
}

// Deprecated: Use UpdateBuildHistoryRequest.ProtoReflect.Descriptor instead.
func (*UpdateBuildHistoryRequest) Descriptor() ([]byte, []int) {
	return file_github_com_moby_buildkit_api_services_control_control_proto_rawDescGZIP(), []int{22}
}

func (x *UpdateBuildHistoryRequest) GetRef() string {
	if x != nil {
		return x.Ref
	}
	return ""
}

func (x *UpdateBuildHistoryRequest) GetPinned() bool {
	if x != nil {
		return x.Pinned
	}
	return false
}

func (x *UpdateBuildHistoryRequest) GetDelete() bool {
	if x != nil {
		return x.Delete
	}
	return false
}

func (x *UpdateBuildHistoryRequest) GetFinalize() bool {
	if x != nil {
		return x.Finalize
	}
	return false
}

type UpdateBuildHistoryResponse struct {
	state         protoimpl.MessageState `protogen:"open.v1"`
	unknownFields protoimpl.UnknownFields
	sizeCache     protoimpl.SizeCache
}

func (x *UpdateBuildHistoryResponse) Reset() {
	*x = UpdateBuildHistoryResponse{}
	mi := &file_github_com_moby_buildkit_api_services_control_control_proto_msgTypes[23]
	ms := protoimpl.X.MessageStateOf(protoimpl.Pointer(x))
	ms.StoreMessageInfo(mi)
}

func (x *UpdateBuildHistoryResponse) String() string {
	return protoimpl.X.MessageStringOf(x)
}

func (*UpdateBuildHistoryResponse) ProtoMessage() {}

func (x *UpdateBuildHistoryResponse) ProtoReflect() protoreflect.Message {
	mi := &file_github_com_moby_buildkit_api_services_control_control_proto_msgTypes[23]
	if x != nil {
		ms := protoimpl.X.MessageStateOf(protoimpl.Pointer(x))
		if ms.LoadMessageInfo() == nil {
			ms.StoreMessageInfo(mi)
		}
		return ms
	}
	return mi.MessageOf(x)
}

// Deprecated: Use UpdateBuildHistoryResponse.ProtoReflect.Descriptor instead.
func (*UpdateBuildHistoryResponse) Descriptor() ([]byte, []int) {
	return file_github_com_moby_buildkit_api_services_control_control_proto_rawDescGZIP(), []int{23}
}

type Descriptor struct {
	state         protoimpl.MessageState `protogen:"open.v1"`
	MediaType     string                 `protobuf:"bytes,1,opt,name=media_type,json=mediaType,proto3" json:"media_type,omitempty"`
	Digest        string                 `protobuf:"bytes,2,opt,name=digest,proto3" json:"digest,omitempty"`
	Size          int64                  `protobuf:"varint,3,opt,name=size,proto3" json:"size,omitempty"`
	Annotations   map[string]string      `protobuf:"bytes,5,rep,name=annotations,proto3" json:"annotations,omitempty" protobuf_key:"bytes,1,opt,name=key" protobuf_val:"bytes,2,opt,name=value"`
	unknownFields protoimpl.UnknownFields
	sizeCache     protoimpl.SizeCache
}

func (x *Descriptor) Reset() {
	*x = Descriptor{}
	mi := &file_github_com_moby_buildkit_api_services_control_control_proto_msgTypes[24]
	ms := protoimpl.X.MessageStateOf(protoimpl.Pointer(x))
	ms.StoreMessageInfo(mi)
}

func (x *Descriptor) String() string {
	return protoimpl.X.MessageStringOf(x)
}

func (*Descriptor) ProtoMessage() {}

func (x *Descriptor) ProtoReflect() protoreflect.Message {
	mi := &file_github_com_moby_buildkit_api_services_control_control_proto_msgTypes[24]
	if x != nil {
		ms := protoimpl.X.MessageStateOf(protoimpl.Pointer(x))
		if ms.LoadMessageInfo() == nil {
			ms.StoreMessageInfo(mi)
		}
		return ms
	}
	return mi.MessageOf(x)
}

// Deprecated: Use Descriptor.ProtoReflect.Descriptor instead.
func (*Descriptor) Descriptor() ([]byte, []int) {
	return file_github_com_moby_buildkit_api_services_control_control_proto_rawDescGZIP(), []int{24}
}

func (x *Descriptor) GetMediaType() string {
	if x != nil {
		return x.MediaType
	}
	return ""
}

func (x *Descriptor) GetDigest() string {
	if x != nil {
		return x.Digest
	}
	return ""
}

func (x *Descriptor) GetSize() int64 {
	if x != nil {
		return x.Size
	}
	return 0
}

func (x *Descriptor) GetAnnotations() map[string]string {
	if x != nil {
		return x.Annotations
	}
	return nil
}

type BuildResultInfo struct {
	state            protoimpl.MessageState `protogen:"open.v1"`
	ResultDeprecated *Descriptor            `protobuf:"bytes,1,opt,name=ResultDeprecated,proto3" json:"ResultDeprecated,omitempty"`
	Attestations     []*Descriptor          `protobuf:"bytes,2,rep,name=Attestations,proto3" json:"Attestations,omitempty"`
	Results          map[int64]*Descriptor  `protobuf:"bytes,3,rep,name=Results,proto3" json:"Results,omitempty" protobuf_key:"varint,1,opt,name=key" protobuf_val:"bytes,2,opt,name=value"`
	unknownFields    protoimpl.UnknownFields
	sizeCache        protoimpl.SizeCache
}

func (x *BuildResultInfo) Reset() {
	*x = BuildResultInfo{}
	mi := &file_github_com_moby_buildkit_api_services_control_control_proto_msgTypes[25]
	ms := protoimpl.X.MessageStateOf(protoimpl.Pointer(x))
	ms.StoreMessageInfo(mi)
}

func (x *BuildResultInfo) String() string {
	return protoimpl.X.MessageStringOf(x)
}

func (*BuildResultInfo) ProtoMessage() {}

func (x *BuildResultInfo) ProtoReflect() protoreflect.Message {
	mi := &file_github_com_moby_buildkit_api_services_control_control_proto_msgTypes[25]
	if x != nil {
		ms := protoimpl.X.MessageStateOf(protoimpl.Pointer(x))
		if ms.LoadMessageInfo() == nil {
			ms.StoreMessageInfo(mi)
		}
		return ms
	}
	return mi.MessageOf(x)
}

// Deprecated: Use BuildResultInfo.ProtoReflect.Descriptor instead.
func (*BuildResultInfo) Descriptor() ([]byte, []int) {
	return file_github_com_moby_buildkit_api_services_control_control_proto_rawDescGZIP(), []int{25}
}

func (x *BuildResultInfo) GetResultDeprecated() *Descriptor {
	if x != nil {
		return x.ResultDeprecated
	}
	return nil
}

func (x *BuildResultInfo) GetAttestations() []*Descriptor {
	if x != nil {
		return x.Attestations
	}
	return nil
}

func (x *BuildResultInfo) GetResults() map[int64]*Descriptor {
	if x != nil {
		return x.Results
	}
	return nil
}

// Exporter describes the output exporter
type Exporter struct {
	state protoimpl.MessageState `protogen:"open.v1"`
	// Type identifies the exporter
	Type string `protobuf:"bytes,1,opt,name=Type,proto3" json:"Type,omitempty"`
	// Attrs specifies exporter configuration
	Attrs         map[string]string `protobuf:"bytes,2,rep,name=Attrs,proto3" json:"Attrs,omitempty" protobuf_key:"bytes,1,opt,name=key" protobuf_val:"bytes,2,opt,name=value"`
	unknownFields protoimpl.UnknownFields
	sizeCache     protoimpl.SizeCache
}

func (x *Exporter) Reset() {
	*x = Exporter{}
	mi := &file_github_com_moby_buildkit_api_services_control_control_proto_msgTypes[26]
	ms := protoimpl.X.MessageStateOf(protoimpl.Pointer(x))
	ms.StoreMessageInfo(mi)
}

func (x *Exporter) String() string {
	return protoimpl.X.MessageStringOf(x)
}

func (*Exporter) ProtoMessage() {}

func (x *Exporter) ProtoReflect() protoreflect.Message {
	mi := &file_github_com_moby_buildkit_api_services_control_control_proto_msgTypes[26]
	if x != nil {
		ms := protoimpl.X.MessageStateOf(protoimpl.Pointer(x))
		if ms.LoadMessageInfo() == nil {
			ms.StoreMessageInfo(mi)
		}
		return ms
	}
	return mi.MessageOf(x)
}

// Deprecated: Use Exporter.ProtoReflect.Descriptor instead.
func (*Exporter) Descriptor() ([]byte, []int) {
	return file_github_com_moby_buildkit_api_services_control_control_proto_rawDescGZIP(), []int{26}
}

func (x *Exporter) GetType() string {
	if x != nil {
		return x.Type
	}
	return ""
}

func (x *Exporter) GetAttrs() map[string]string {
	if x != nil {
		return x.Attrs
	}
	return nil
}

var File_github_com_moby_buildkit_api_services_control_control_proto protoreflect.FileDescriptor

const file_github_com_moby_buildkit_api_services_control_control_proto_rawDesc = "" +
	"\n" +
	";github.com/moby/buildkit/api/services/control/control.proto\x12\x10moby.buildkit.v1\x1a/github.com/moby/buildkit/api/types/worker.proto\x1a,github.com/moby/buildkit/solver/pb/ops.proto\x1a5github.com/moby/buildkit/sourcepolicy/pb/policy.proto\x1a\x1fgoogle/protobuf/timestamp.proto\x1a\x17google/rpc/status.proto\"\xca\x01\n" +
	"\fPruneRequest\x12\x16\n" +
	"\x06filter\x18\x01 \x03(\tR\x06filter\x12\x10\n" +
	"\x03all\x18\x02 \x01(\bR\x03all\x12\"\n" +
	"\fkeepDuration\x18\x03 \x01(\x03R\fkeepDuration\x12$\n" +
	"\rreservedSpace\x18\x04 \x01(\x03R\rreservedSpace\x12\"\n" +
	"\fmaxUsedSpace\x18\x05 \x01(\x03R\fmaxUsedSpace\x12\"\n" +
	"\fminFreeSpace\x18\x06 \x01(\x03R\fminFreeSpace\"F\n" +
	"\x10DiskUsageRequest\x12\x16\n" +
	"\x06filter\x18\x01 \x03(\tR\x06filter\x12\x1a\n" +
	"\bageLimit\x18\x02 \x01(\x03R\bageLimit\"J\n" +
	"\x11DiskUsageResponse\x125\n" +
	"\x06record\x18\x01 \x03(\v2\x1d.moby.buildkit.v1.UsageRecordR\x06record\"\x87\x03\n" +
	"\vUsageRecord\x12\x0e\n" +
	"\x02ID\x18\x01 \x01(\tR\x02ID\x12\x18\n" +
	"\aMutable\x18\x02 \x01(\bR\aMutable\x12\x14\n" +
	"\x05InUse\x18\x03 \x01(\bR\x05InUse\x12\x12\n" +
	"\x04Size\x18\x04 \x01(\x03R\x04Size\x12\x1a\n" +
	"\x06Parent\x18\x05 \x01(\tB\x02\x18\x01R\x06Parent\x128\n" +
	"\tCreatedAt\x18\x06 \x01(\v2\x1a.google.protobuf.TimestampR\tCreatedAt\x12:\n" +
	"\n" +
	"LastUsedAt\x18\a \x01(\v2\x1a.google.protobuf.TimestampR\n" +
	"LastUsedAt\x12\x1e\n" +
	"\n" +
	"UsageCount\x18\b \x01(\x03R\n" +
	"UsageCount\x12 \n" +
	"\vDescription\x18\t \x01(\tR\vDescription\x12\x1e\n" +
	"\n" +
	"RecordType\x18\n" +
	" \x01(\tR\n" +
	"RecordType\x12\x16\n" +
	"\x06Shared\x18\v \x01(\bR\x06Shared\x12\x18\n" +
	"\aParents\x18\f \x03(\tR\aParents\"\xf4\a\n" +
	"\fSolveRequest\x12\x10\n" +
	"\x03Ref\x18\x01 \x01(\tR\x03Ref\x12.\n" +
	"\n" +
	"Definition\x18\x02 \x01(\v2\x0e.pb.DefinitionR\n" +
	"Definition\x12.\n" +
	"\x12ExporterDeprecated\x18\x03 \x01(\tR\x12ExporterDeprecated\x12u\n" +
	"\x17ExporterAttrsDeprecated\x18\x04 \x03(\v2;.moby.buildkit.v1.SolveRequest.ExporterAttrsDeprecatedEntryR\x17ExporterAttrsDeprecated\x12\x18\n" +
	"\aSession\x18\x05 \x01(\tR\aSession\x12\x1a\n" +
	"\bFrontend\x18\x06 \x01(\tR\bFrontend\x12W\n" +
	"\rFrontendAttrs\x18\a \x03(\v21.moby.buildkit.v1.SolveRequest.FrontendAttrsEntryR\rFrontendAttrs\x124\n" +
	"\x05Cache\x18\b \x01(\v2\x1e.moby.buildkit.v1.CacheOptionsR\x05Cache\x12\"\n" +
	"\fEntitlements\x18\t \x03(\tR\fEntitlements\x12Z\n" +
	"\x0eFrontendInputs\x18\n" +
	" \x03(\v22.moby.buildkit.v1.SolveRequest.FrontendInputsEntryR\x0eFrontendInputs\x12\x1a\n" +
	"\bInternal\x18\v \x01(\bR\bInternal\x12I\n" +
	"\fSourcePolicy\x18\f \x01(\v2%.moby.buildkit.v1.sourcepolicy.PolicyR\fSourcePolicy\x128\n" +
	"\tExporters\x18\r \x03(\v2\x1a.moby.buildkit.v1.ExporterR\tExporters\x124\n" +
	"\x15EnableSessionExporter\x18\x0e \x01(\bR\x15EnableSessionExporter\x1aJ\n" +
	"\x1cExporterAttrsDeprecatedEntry\x12\x10\n" +
	"\x03key\x18\x01 \x01(\tR\x03key\x12\x14\n" +
	"\x05value\x18\x02 \x01(\tR\x05value:\x028\x01\x1a@\n" +
	"\x12FrontendAttrsEntry\x12\x10\n" +
	"\x03key\x18\x01 \x01(\tR\x03key\x12\x14\n" +
	"\x05value\x18\x02 \x01(\tR\x05value:\x028\x01\x1aQ\n" +
	"\x13FrontendInputsEntry\x12\x10\n" +
	"\x03key\x18\x01 \x01(\tR\x03key\x12$\n" +
	"\x05value\x18\x02 \x01(\v2\x0e.pb.DefinitionR\x05value:\x028\x01\"\xad\x03\n" +
	"\fCacheOptions\x120\n" +
	"\x13ExportRefDeprecated\x18\x01 \x01(\tR\x13ExportRefDeprecated\x122\n" +
	"\x14ImportRefsDeprecated\x18\x02 \x03(\tR\x14ImportRefsDeprecated\x12o\n" +
	"\x15ExportAttrsDeprecated\x18\x03 \x03(\v29.moby.buildkit.v1.CacheOptions.ExportAttrsDeprecatedEntryR\x15ExportAttrsDeprecated\x12=\n" +
	"\aExports\x18\x04 \x03(\v2#.moby.buildkit.v1.CacheOptionsEntryR\aExports\x12=\n" +
	"\aImports\x18\x05 \x03(\v2#.moby.buildkit.v1.CacheOptionsEntryR\aImports\x1aH\n" +
	"\x1aExportAttrsDeprecatedEntry\x12\x10\n" +
	"\x03key\x18\x01 \x01(\tR\x03key\x12\x14\n" +
	"\x05value\x18\x02 \x01(\tR\x05value:\x028\x01\"\xa7\x01\n" +
	"\x11CacheOptionsEntry\x12\x12\n" +
	"\x04Type\x18\x01 \x01(\tR\x04Type\x12D\n" +
	"\x05Attrs\x18\x02 \x03(\v2..moby.buildkit.v1.CacheOptionsEntry.AttrsEntryR\x05Attrs\x1a8\n" +
	"\n" +
	"AttrsEntry\x12\x10\n" +
	"\x03key\x18\x01 \x01(\tR\x03key\x12\x14\n" +
	"\x05value\x18\x02 \x01(\tR\x05value:\x028\x01\"\xb7\x01\n" +
	"\rSolveResponse\x12a\n" +
	"\x10ExporterResponse\x18\x01 \x03(\v25.moby.buildkit.v1.SolveResponse.ExporterResponseEntryR\x10ExporterResponse\x1aC\n" +
	"\x15ExporterResponseEntry\x12\x10\n" +
	"\x03key\x18\x01 \x01(\tR\x03key\x12\x14\n" +
	"\x05value\x18\x02 \x01(\tR\x05value:\x028\x01\"!\n" +
	"\rStatusRequest\x12\x10\n" +
	"\x03Ref\x18\x01 \x01(\tR\x03Ref\"\xf0\x01\n" +
	"\x0eStatusResponse\x124\n" +
	"\bvertexes\x18\x01 \x03(\v2\x18.moby.buildkit.v1.VertexR\bvertexes\x12:\n" +
	"\bstatuses\x18\x02 \x03(\v2\x1e.moby.buildkit.v1.VertexStatusR\bstatuses\x12/\n" +
	"\x04logs\x18\x03 \x03(\v2\x1b.moby.buildkit.v1.VertexLogR\x04logs\x12;\n" +
	"\bwarnings\x18\x04 \x03(\v2\x1f.moby.buildkit.v1.VertexWarningR\bwarnings\"\xa3\x02\n" +
	"\x06Vertex\x12\x16\n" +
	"\x06digest\x18\x01 \x01(\tR\x06digest\x12\x16\n" +
	"\x06inputs\x18\x02 \x03(\tR\x06inputs\x12\x12\n" +
	"\x04name\x18\x03 \x01(\tR\x04name\x12\x16\n" +
	"\x06cached\x18\x04 \x01(\bR\x06cached\x124\n" +
	"\astarted\x18\x05 \x01(\v2\x1a.google.protobuf.TimestampR\astarted\x128\n" +
	"\tcompleted\x18\x06 \x01(\v2\x1a.google.protobuf.TimestampR\tcompleted\x12\x14\n" +
	"\x05error\x18\a \x01(\tR\x05error\x127\n" +
	"\rprogressGroup\x18\b \x01(\v2\x11.pb.ProgressGroupR\rprogressGroup\"\xa4\x02\n" +
	"\fVertexStatus\x12\x0e\n" +
	"\x02ID\x18\x01 \x01(\tR\x02ID\x12\x16\n" +
	"\x06vertex\x18\x02 \x01(\tR\x06vertex\x12\x12\n" +
	"\x04name\x18\x03 \x01(\tR\x04name\x12\x18\n" +
	"\acurrent\x18\x04 \x01(\x03R\acurrent\x12\x14\n" +
	"\x05total\x18\x05 \x01(\x03R\x05total\x128\n" +
	"\ttimestamp\x18\x06 \x01(\v2\x1a.google.protobuf.TimestampR\ttimestamp\x124\n" +
	"\astarted\x18\a \x01(\v2\x1a.google.protobuf.TimestampR\astarted\x128\n" +
	"\tcompleted\x18\b \x01(\v2\x1a.google.protobuf.TimestampR\tcompleted\"\x87\x01\n" +
	"\tVertexLog\x12\x16\n" +
	"\x06vertex\x18\x01 \x01(\tR\x06vertex\x128\n" +
	"\ttimestamp\x18\x02 \x01(\v2\x1a.google.protobuf.TimestampR\ttimestamp\x12\x16\n" +
	"\x06stream\x18\x03 \x01(\x03R\x06stream\x12\x10\n" +
	"\x03msg\x18\x04 \x01(\fR\x03msg\"\xc4\x01\n" +
	"\rVertexWarning\x12\x16\n" +
	"\x06vertex\x18\x01 \x01(\tR\x06vertex\x12\x14\n" +
	"\x05level\x18\x02 \x01(\x03R\x05level\x12\x14\n" +
	"\x05short\x18\x03 \x01(\fR\x05short\x12\x16\n" +
	"\x06detail\x18\x04 \x03(\fR\x06detail\x12\x10\n" +
	"\x03url\x18\x05 \x01(\tR\x03url\x12\"\n" +
	"\x04info\x18\x06 \x01(\v2\x0e.pb.SourceInfoR\x04info\x12!\n" +
	"\x06ranges\x18\a \x03(\v2\t.pb.RangeR\x06ranges\"\"\n" +
	"\fBytesMessage\x12\x12\n" +
	"\x04data\x18\x01 \x01(\fR\x04data\",\n" +
	"\x12ListWorkersRequest\x12\x16\n" +
	"\x06filter\x18\x01 \x03(\tR\x06filter\"S\n" +
	"\x13ListWorkersResponse\x12<\n" +
	"\x06record\x18\x01 \x03(\v2$.moby.buildkit.v1.types.WorkerRecordR\x06record\"\r\n" +
	"\vInfoRequest\"a\n" +
	"\fInfoResponse\x12Q\n" +
	"\x0fbuildkitVersion\x18\x01 \x01(\v2'.moby.buildkit.v1.types.BuildkitVersionR\x0fbuildkitVersion\"\x93\x01\n" +
	"\x13BuildHistoryRequest\x12\x1e\n" +
	"\n" +
	"ActiveOnly\x18\x01 \x01(\bR\n" +
	"ActiveOnly\x12\x10\n" +
	"\x03Ref\x18\x02 \x01(\tR\x03Ref\x12\x1c\n" +
	"\tEarlyExit\x18\x03 \x01(\bR\tEarlyExit\x12\x16\n" +
	"\x06Filter\x18\x04 \x03(\tR\x06Filter\x12\x14\n" +
	"\x05Limit\x18\x05 \x01(\x05R\x05Limit\"\x8e\x01\n" +
	"\x11BuildHistoryEvent\x12;\n" +
	"\x04type\x18\x01 \x01(\x0e2'.moby.buildkit.v1.BuildHistoryEventTypeR\x04type\x12<\n" +
	"\x06record\x18\x02 \x01(\v2$.moby.buildkit.v1.BuildHistoryRecordR\x06record\"\xd3\t\n" +
	"\x12BuildHistoryRecord\x12\x10\n" +
	"\x03Ref\x18\x01 \x01(\tR\x03Ref\x12\x1a\n" +
	"\bFrontend\x18\x02 \x01(\tR\bFrontend\x12]\n" +
	"\rFrontendAttrs\x18\x03 \x03(\v27.moby.buildkit.v1.BuildHistoryRecord.FrontendAttrsEntryR\rFrontendAttrs\x128\n" +
	"\tExporters\x18\x04 \x03(\v2\x1a.moby.buildkit.v1.ExporterR\tExporters\x12(\n" +
	"\x05error\x18\x05 \x01(\v2\x12.google.rpc.StatusR\x05error\x128\n" +
	"\tCreatedAt\x18\x06 \x01(\v2\x1a.google.protobuf.TimestampR\tCreatedAt\x12<\n" +
	"\vCompletedAt\x18\a \x01(\v2\x1a.google.protobuf.TimestampR\vCompletedAt\x120\n" +
	"\x04logs\x18\b \x01(\v2\x1c.moby.buildkit.v1.DescriptorR\x04logs\x12f\n" +
	"\x10ExporterResponse\x18\t \x03(\v2:.moby.buildkit.v1.BuildHistoryRecord.ExporterResponseEntryR\x10ExporterResponse\x129\n" +
	"\x06Result\x18\n" +
	" \x01(\v2!.moby.buildkit.v1.BuildResultInfoR\x06Result\x12K\n" +
	"\aResults\x18\v \x03(\v21.moby.buildkit.v1.BuildHistoryRecord.ResultsEntryR\aResults\x12\x1e\n" +
	"\n" +
	"Generation\x18\f \x01(\x05R\n" +
	"Generation\x122\n" +
	"\x05trace\x18\r \x01(\v2\x1c.moby.buildkit.v1.DescriptorR\x05trace\x12\x16\n" +
	"\x06pinned\x18\x0e \x01(\bR\x06pinned\x12&\n" +
	"\x0enumCachedSteps\x18\x0f \x01(\x05R\x0enumCachedSteps\x12$\n" +
	"\rnumTotalSteps\x18\x10 \x01(\x05R\rnumTotalSteps\x12,\n" +
	"\x11numCompletedSteps\x18\x11 \x01(\x05R\x11numCompletedSteps\x12B\n" +
	"\rexternalError\x18\x12 \x01(\v2\x1c.moby.buildkit.v1.DescriptorR\rexternalError\x12 \n" +
	"\vnumWarnings\x18\x13 \x01(\x05R\vnumWarnings\x1a@\n" +
	"\x12FrontendAttrsEntry\x12\x10\n" +
	"\x03key\x18\x01 \x01(\tR\x03key\x12\x14\n" +
	"\x05value\x18\x02 \x01(\tR\x05value:\x028\x01\x1aC\n" +
	"\x15ExporterResponseEntry\x12\x10\n" +
	"\x03key\x18\x01 \x01(\tR\x03key\x12\x14\n" +
	"\x05value\x18\x02 \x01(\tR\x05value:\x028\x01\x1a]\n" +
	"\fResultsEntry\x12\x10\n" +
	"\x03key\x18\x01 \x01(\tR\x03key\x127\n" +
	"\x05value\x18\x02 \x01(\v2!.moby.buildkit.v1.BuildResultInfoR\x05value:\x028\x01\"y\n" +
	"\x19UpdateBuildHistoryRequest\x12\x10\n" +
	"\x03Ref\x18\x01 \x01(\tR\x03Ref\x12\x16\n" +
	"\x06Pinned\x18\x02 \x01(\bR\x06Pinned\x12\x16\n" +
	"\x06Delete\x18\x03 \x01(\bR\x06Delete\x12\x1a\n" +
	"\bFinalize\x18\x04 \x01(\bR\bFinalize\"\x1c\n" +
	"\x1aUpdateBuildHistoryResponse\"\xe8\x01\n" +
	"\n" +
	"Descriptor\x12\x1d\n" +
	"\n" +
	"media_type\x18\x01 \x01(\tR\tmediaType\x12\x16\n" +
	"\x06digest\x18\x02 \x01(\tR\x06digest\x12\x12\n" +
	"\x04size\x18\x03 \x01(\x03R\x04size\x12O\n" +
	"\vannotations\x18\x05 \x03(\v2-.moby.buildkit.v1.Descriptor.AnnotationsEntryR\vannotations\x1a>\n" +
	"\x10AnnotationsEntry\x12\x10\n" +
	"\x03key\x18\x01 \x01(\tR\x03key\x12\x14\n" +
	"\x05value\x18\x02 \x01(\tR\x05value:\x028\x01\"\xc1\x02\n" +
	"\x0fBuildResultInfo\x12H\n" +
	"\x10ResultDeprecated\x18\x01 \x01(\v2\x1c.moby.buildkit.v1.DescriptorR\x10ResultDeprecated\x12@\n" +
	"\fAttestations\x18\x02 \x03(\v2\x1c.moby.buildkit.v1.DescriptorR\fAttestations\x12H\n" +
	"\aResults\x18\x03 \x03(\v2..moby.buildkit.v1.BuildResultInfo.ResultsEntryR\aResults\x1aX\n" +
	"\fResultsEntry\x12\x10\n" +
	"\x03key\x18\x01 \x01(\x03R\x03key\x122\n" +
	"\x05value\x18\x02 \x01(\v2\x1c.moby.buildkit.v1.DescriptorR\x05value:\x028\x01\"\x95\x01\n" +
	"\bExporter\x12\x12\n" +
	"\x04Type\x18\x01 \x01(\tR\x04Type\x12;\n" +
	"\x05Attrs\x18\x02 \x03(\v2%.moby.buildkit.v1.Exporter.AttrsEntryR\x05Attrs\x1a8\n" +
	"\n" +
	"AttrsEntry\x12\x10\n" +
	"\x03key\x18\x01 \x01(\tR\x03key\x12\x14\n" +
	"\x05value\x18\x02 \x01(\tR\x05value:\x028\x01*?\n" +
	"\x15BuildHistoryEventType\x12\v\n" +
	"\aSTARTED\x10\x00\x12\f\n" +
	"\bCOMPLETE\x10\x01\x12\v\n" +
	"\aDELETED\x10\x022\x89\x06\n" +
	"\aControl\x12T\n" +
	"\tDiskUsage\x12\".moby.buildkit.v1.DiskUsageRequest\x1a#.moby.buildkit.v1.DiskUsageResponse\x12H\n" +
	"\x05Prune\x12\x1e.moby.buildkit.v1.PruneRequest\x1a\x1d.moby.buildkit.v1.UsageRecord0\x01\x12H\n" +
	"\x05Solve\x12\x1e.moby.buildkit.v1.SolveRequest\x1a\x1f.moby.buildkit.v1.SolveResponse\x12M\n" +
	"\x06Status\x12\x1f.moby.buildkit.v1.StatusRequest\x1a .moby.buildkit.v1.StatusResponse0\x01\x12M\n" +
	"\aSession\x12\x1e.moby.buildkit.v1.BytesMessage\x1a\x1e.moby.buildkit.v1.BytesMessage(\x010\x01\x12Z\n" +
	"\vListWorkers\x12$.moby.buildkit.v1.ListWorkersRequest\x1a%.moby.buildkit.v1.ListWorkersResponse\x12E\n" +
	"\x04Info\x12\x1d.moby.buildkit.v1.InfoRequest\x1a\x1e.moby.buildkit.v1.InfoResponse\x12b\n" +
	"\x12ListenBuildHistory\x12%.moby.buildkit.v1.BuildHistoryRequest\x1a#.moby.buildkit.v1.BuildHistoryEvent0\x01\x12o\n" +
	"\x12UpdateBuildHistory\x12+.moby.buildkit.v1.UpdateBuildHistoryRequest\x1a,.moby.buildkit.v1.UpdateBuildHistoryResponseB@Z>github.com/moby/buildkit/api/services/control;moby_buildkit_v1b\x06proto3"

var (
	file_github_com_moby_buildkit_api_services_control_control_proto_rawDescOnce sync.Once
	file_github_com_moby_buildkit_api_services_control_control_proto_rawDescData []byte
)

func file_github_com_moby_buildkit_api_services_control_control_proto_rawDescGZIP() []byte {
	file_github_com_moby_buildkit_api_services_control_control_proto_rawDescOnce.Do(func() {
		file_github_com_moby_buildkit_api_services_control_control_proto_rawDescData = protoimpl.X.CompressGZIP(unsafe.Slice(unsafe.StringData(file_github_com_moby_buildkit_api_services_control_control_proto_rawDesc), len(file_github_com_moby_buildkit_api_services_control_control_proto_rawDesc)))
	})
	return file_github_com_moby_buildkit_api_services_control_control_proto_rawDescData
}

var file_github_com_moby_buildkit_api_services_control_control_proto_enumTypes = make([]protoimpl.EnumInfo, 1)
var file_github_com_moby_buildkit_api_services_control_control_proto_msgTypes = make([]protoimpl.MessageInfo, 39)
var file_github_com_moby_buildkit_api_services_control_control_proto_goTypes = []any{
	(BuildHistoryEventType)(0),         // 0: moby.buildkit.v1.BuildHistoryEventType
	(*PruneRequest)(nil),               // 1: moby.buildkit.v1.PruneRequest
	(*DiskUsageRequest)(nil),           // 2: moby.buildkit.v1.DiskUsageRequest
	(*DiskUsageResponse)(nil),          // 3: moby.buildkit.v1.DiskUsageResponse
	(*UsageRecord)(nil),                // 4: moby.buildkit.v1.UsageRecord
	(*SolveRequest)(nil),               // 5: moby.buildkit.v1.SolveRequest
	(*CacheOptions)(nil),               // 6: moby.buildkit.v1.CacheOptions
	(*CacheOptionsEntry)(nil),          // 7: moby.buildkit.v1.CacheOptionsEntry
	(*SolveResponse)(nil),              // 8: moby.buildkit.v1.SolveResponse
	(*StatusRequest)(nil),              // 9: moby.buildkit.v1.StatusRequest
	(*StatusResponse)(nil),             // 10: moby.buildkit.v1.StatusResponse
	(*Vertex)(nil),                     // 11: moby.buildkit.v1.Vertex
	(*VertexStatus)(nil),               // 12: moby.buildkit.v1.VertexStatus
	(*VertexLog)(nil),                  // 13: moby.buildkit.v1.VertexLog
	(*VertexWarning)(nil),              // 14: moby.buildkit.v1.VertexWarning
	(*BytesMessage)(nil),               // 15: moby.buildkit.v1.BytesMessage
	(*ListWorkersRequest)(nil),         // 16: moby.buildkit.v1.ListWorkersRequest
	(*ListWorkersResponse)(nil),        // 17: moby.buildkit.v1.ListWorkersResponse
	(*InfoRequest)(nil),                // 18: moby.buildkit.v1.InfoRequest
	(*InfoResponse)(nil),               // 19: moby.buildkit.v1.InfoResponse
	(*BuildHistoryRequest)(nil),        // 20: moby.buildkit.v1.BuildHistoryRequest
	(*BuildHistoryEvent)(nil),          // 21: moby.buildkit.v1.BuildHistoryEvent
	(*BuildHistoryRecord)(nil),         // 22: moby.buildkit.v1.BuildHistoryRecord
	(*UpdateBuildHistoryRequest)(nil),  // 23: moby.buildkit.v1.UpdateBuildHistoryRequest
	(*UpdateBuildHistoryResponse)(nil), // 24: moby.buildkit.v1.UpdateBuildHistoryResponse
	(*Descriptor)(nil),                 // 25: moby.buildkit.v1.Descriptor
	(*BuildResultInfo)(nil),            // 26: moby.buildkit.v1.BuildResultInfo
	(*Exporter)(nil),                   // 27: moby.buildkit.v1.Exporter
	nil,                                // 28: moby.buildkit.v1.SolveRequest.ExporterAttrsDeprecatedEntry
	nil,                                // 29: moby.buildkit.v1.SolveRequest.FrontendAttrsEntry
	nil,                                // 30: moby.buildkit.v1.SolveRequest.FrontendInputsEntry
	nil,                                // 31: moby.buildkit.v1.CacheOptions.ExportAttrsDeprecatedEntry
	nil,                                // 32: moby.buildkit.v1.CacheOptionsEntry.AttrsEntry
	nil,                                // 33: moby.buildkit.v1.SolveResponse.ExporterResponseEntry
	nil,                                // 34: moby.buildkit.v1.BuildHistoryRecord.FrontendAttrsEntry
	nil,                                // 35: moby.buildkit.v1.BuildHistoryRecord.ExporterResponseEntry
	nil,                                // 36: moby.buildkit.v1.BuildHistoryRecord.ResultsEntry
	nil,                                // 37: moby.buildkit.v1.Descriptor.AnnotationsEntry
	nil,                                // 38: moby.buildkit.v1.BuildResultInfo.ResultsEntry
	nil,                                // 39: moby.buildkit.v1.Exporter.AttrsEntry
	(*timestamp.Timestamp)(nil),        // 40: google.protobuf.Timestamp
	(*pb.Definition)(nil),              // 41: pb.Definition
	(*pb1.Policy)(nil),                 // 42: moby.buildkit.v1.sourcepolicy.Policy
	(*pb.ProgressGroup)(nil),           // 43: pb.ProgressGroup
	(*pb.SourceInfo)(nil),              // 44: pb.SourceInfo
	(*pb.Range)(nil),                   // 45: pb.Range
	(*types.WorkerRecord)(nil),         // 46: moby.buildkit.v1.types.WorkerRecord
	(*types.BuildkitVersion)(nil),      // 47: moby.buildkit.v1.types.BuildkitVersion
	(*status.Status)(nil),              // 48: google.rpc.Status
}
var file_github_com_moby_buildkit_api_services_control_control_proto_depIdxs = []int32{
	4,  // 0: moby.buildkit.v1.DiskUsageResponse.record:type_name -> moby.buildkit.v1.UsageRecord
	40, // 1: moby.buildkit.v1.UsageRecord.CreatedAt:type_name -> google.protobuf.Timestamp
	40, // 2: moby.buildkit.v1.UsageRecord.LastUsedAt:type_name -> google.protobuf.Timestamp
	41, // 3: moby.buildkit.v1.SolveRequest.Definition:type_name -> pb.Definition
	28, // 4: moby.buildkit.v1.SolveRequest.ExporterAttrsDeprecated:type_name -> moby.buildkit.v1.SolveRequest.ExporterAttrsDeprecatedEntry
	29, // 5: moby.buildkit.v1.SolveRequest.FrontendAttrs:type_name -> moby.buildkit.v1.SolveRequest.FrontendAttrsEntry
	6,  // 6: moby.buildkit.v1.SolveRequest.Cache:type_name -> moby.buildkit.v1.CacheOptions
	30, // 7: moby.buildkit.v1.SolveRequest.FrontendInputs:type_name -> moby.buildkit.v1.SolveRequest.FrontendInputsEntry
	42, // 8: moby.buildkit.v1.SolveRequest.SourcePolicy:type_name -> moby.buildkit.v1.sourcepolicy.Policy
	27, // 9: moby.buildkit.v1.SolveRequest.Exporters:type_name -> moby.buildkit.v1.Exporter
	31, // 10: moby.buildkit.v1.CacheOptions.ExportAttrsDeprecated:type_name -> moby.buildkit.v1.CacheOptions.ExportAttrsDeprecatedEntry
	7,  // 11: moby.buildkit.v1.CacheOptions.Exports:type_name -> moby.buildkit.v1.CacheOptionsEntry
	7,  // 12: moby.buildkit.v1.CacheOptions.Imports:type_name -> moby.buildkit.v1.CacheOptionsEntry
	32, // 13: moby.buildkit.v1.CacheOptionsEntry.Attrs:type_name -> moby.buildkit.v1.CacheOptionsEntry.AttrsEntry
	33, // 14: moby.buildkit.v1.SolveResponse.ExporterResponse:type_name -> moby.buildkit.v1.SolveResponse.ExporterResponseEntry
	11, // 15: moby.buildkit.v1.StatusResponse.vertexes:type_name -> moby.buildkit.v1.Vertex
	12, // 16: moby.buildkit.v1.StatusResponse.statuses:type_name -> moby.buildkit.v1.VertexStatus
	13, // 17: moby.buildkit.v1.StatusResponse.logs:type_name -> moby.buildkit.v1.VertexLog
	14, // 18: moby.buildkit.v1.StatusResponse.warnings:type_name -> moby.buildkit.v1.VertexWarning
	40, // 19: moby.buildkit.v1.Vertex.started:type_name -> google.protobuf.Timestamp
	40, // 20: moby.buildkit.v1.Vertex.completed:type_name -> google.protobuf.Timestamp
	43, // 21: moby.buildkit.v1.Vertex.progressGroup:type_name -> pb.ProgressGroup
	40, // 22: moby.buildkit.v1.VertexStatus.timestamp:type_name -> google.protobuf.Timestamp
	40, // 23: moby.buildkit.v1.VertexStatus.started:type_name -> google.protobuf.Timestamp
	40, // 24: moby.buildkit.v1.VertexStatus.completed:type_name -> google.protobuf.Timestamp
	40, // 25: moby.buildkit.v1.VertexLog.timestamp:type_name -> google.protobuf.Timestamp
	44, // 26: moby.buildkit.v1.VertexWarning.info:type_name -> pb.SourceInfo
	45, // 27: moby.buildkit.v1.VertexWarning.ranges:type_name -> pb.Range
	46, // 28: moby.buildkit.v1.ListWorkersResponse.record:type_name -> moby.buildkit.v1.types.WorkerRecord
	47, // 29: moby.buildkit.v1.InfoResponse.buildkitVersion:type_name -> moby.buildkit.v1.types.BuildkitVersion
	0,  // 30: moby.buildkit.v1.BuildHistoryEvent.type:type_name -> moby.buildkit.v1.BuildHistoryEventType
	22, // 31: moby.buildkit.v1.BuildHistoryEvent.record:type_name -> moby.buildkit.v1.BuildHistoryRecord
	34, // 32: moby.buildkit.v1.BuildHistoryRecord.FrontendAttrs:type_name -> moby.buildkit.v1.BuildHistoryRecord.FrontendAttrsEntry
	27, // 33: moby.buildkit.v1.BuildHistoryRecord.Exporters:type_name -> moby.buildkit.v1.Exporter
	48, // 34: moby.buildkit.v1.BuildHistoryRecord.error:type_name -> google.rpc.Status
	40, // 35: moby.buildkit.v1.BuildHistoryRecord.CreatedAt:type_name -> google.protobuf.Timestamp
	40, // 36: moby.buildkit.v1.BuildHistoryRecord.CompletedAt:type_name -> google.protobuf.Timestamp
	25, // 37: moby.buildkit.v1.BuildHistoryRecord.logs:type_name -> moby.buildkit.v1.Descriptor
	35, // 38: moby.buildkit.v1.BuildHistoryRecord.ExporterResponse:type_name -> moby.buildkit.v1.BuildHistoryRecord.ExporterResponseEntry
	26, // 39: moby.buildkit.v1.BuildHistoryRecord.Result:type_name -> moby.buildkit.v1.BuildResultInfo
	36, // 40: moby.buildkit.v1.BuildHistoryRecord.Results:type_name -> moby.buildkit.v1.BuildHistoryRecord.ResultsEntry
	25, // 41: moby.buildkit.v1.BuildHistoryRecord.trace:type_name -> moby.buildkit.v1.Descriptor
	25, // 42: moby.buildkit.v1.BuildHistoryRecord.externalError:type_name -> moby.buildkit.v1.Descriptor
	37, // 43: moby.buildkit.v1.Descriptor.annotations:type_name -> moby.buildkit.v1.Descriptor.AnnotationsEntry
	25, // 44: moby.buildkit.v1.BuildResultInfo.ResultDeprecated:type_name -> moby.buildkit.v1.Descriptor
	25, // 45: moby.buildkit.v1.BuildResultInfo.Attestations:type_name -> moby.buildkit.v1.Descriptor
	38, // 46: moby.buildkit.v1.BuildResultInfo.Results:type_name -> moby.buildkit.v1.BuildResultInfo.ResultsEntry
	39, // 47: moby.buildkit.v1.Exporter.Attrs:type_name -> moby.buildkit.v1.Exporter.AttrsEntry
	41, // 48: moby.buildkit.v1.SolveRequest.FrontendInputsEntry.value:type_name -> pb.Definition
	26, // 49: moby.buildkit.v1.BuildHistoryRecord.ResultsEntry.value:type_name -> moby.buildkit.v1.BuildResultInfo
	25, // 50: moby.buildkit.v1.BuildResultInfo.ResultsEntry.value:type_name -> moby.buildkit.v1.Descriptor
	2,  // 51: moby.buildkit.v1.Control.DiskUsage:input_type -> moby.buildkit.v1.DiskUsageRequest
	1,  // 52: moby.buildkit.v1.Control.Prune:input_type -> moby.buildkit.v1.PruneRequest
	5,  // 53: moby.buildkit.v1.Control.Solve:input_type -> moby.buildkit.v1.SolveRequest
	9,  // 54: moby.buildkit.v1.Control.Status:input_type -> moby.buildkit.v1.StatusRequest
	15, // 55: moby.buildkit.v1.Control.Session:input_type -> moby.buildkit.v1.BytesMessage
	16, // 56: moby.buildkit.v1.Control.ListWorkers:input_type -> moby.buildkit.v1.ListWorkersRequest
	18, // 57: moby.buildkit.v1.Control.Info:input_type -> moby.buildkit.v1.InfoRequest
	20, // 58: moby.buildkit.v1.Control.ListenBuildHistory:input_type -> moby.buildkit.v1.BuildHistoryRequest
	23, // 59: moby.buildkit.v1.Control.UpdateBuildHistory:input_type -> moby.buildkit.v1.UpdateBuildHistoryRequest
	3,  // 60: moby.buildkit.v1.Control.DiskUsage:output_type -> moby.buildkit.v1.DiskUsageResponse
	4,  // 61: moby.buildkit.v1.Control.Prune:output_type -> moby.buildkit.v1.UsageRecord
	8,  // 62: moby.buildkit.v1.Control.Solve:output_type -> moby.buildkit.v1.SolveResponse
	10, // 63: moby.buildkit.v1.Control.Status:output_type -> moby.buildkit.v1.StatusResponse
	15, // 64: moby.buildkit.v1.Control.Session:output_type -> moby.buildkit.v1.BytesMessage
	17, // 65: moby.buildkit.v1.Control.ListWorkers:output_type -> moby.buildkit.v1.ListWorkersResponse
	19, // 66: moby.buildkit.v1.Control.Info:output_type -> moby.buildkit.v1.InfoResponse
	21, // 67: moby.buildkit.v1.Control.ListenBuildHistory:output_type -> moby.buildkit.v1.BuildHistoryEvent
	24, // 68: moby.buildkit.v1.Control.UpdateBuildHistory:output_type -> moby.buildkit.v1.UpdateBuildHistoryResponse
	60, // [60:69] is the sub-list for method output_type
	51, // [51:60] is the sub-list for method input_type
	51, // [51:51] is the sub-list for extension type_name
	51, // [51:51] is the sub-list for extension extendee
	0,  // [0:51] is the sub-list for field type_name
}

func init() { file_github_com_moby_buildkit_api_services_control_control_proto_init() }
func file_github_com_moby_buildkit_api_services_control_control_proto_init() {
	if File_github_com_moby_buildkit_api_services_control_control_proto != nil {
		return
	}
	type x struct{}
	out := protoimpl.TypeBuilder{
		File: protoimpl.DescBuilder{
			GoPackagePath: reflect.TypeOf(x{}).PkgPath(),
			RawDescriptor: unsafe.Slice(unsafe.StringData(file_github_com_moby_buildkit_api_services_control_control_proto_rawDesc), len(file_github_com_moby_buildkit_api_services_control_control_proto_rawDesc)),
			NumEnums:      1,
			NumMessages:   39,
			NumExtensions: 0,
			NumServices:   1,
		},
		GoTypes:           file_github_com_moby_buildkit_api_services_control_control_proto_goTypes,
		DependencyIndexes: file_github_com_moby_buildkit_api_services_control_control_proto_depIdxs,
		EnumInfos:         file_github_com_moby_buildkit_api_services_control_control_proto_enumTypes,
		MessageInfos:      file_github_com_moby_buildkit_api_services_control_control_proto_msgTypes,
	}.Build()
	File_github_com_moby_buildkit_api_services_control_control_proto = out.File
	file_github_com_moby_buildkit_api_services_control_control_proto_goTypes = nil
	file_github_com_moby_buildkit_api_services_control_control_proto_depIdxs = nil
}
